"""Definition of HDF5 packer plugin interface."""
from __future__ import annotations

from abc import ABC, abstractmethod
from io import UnsupportedOperation
from pathlib import Path
from typing import Callable, Tuple, Type

import wrapt
from overrides import EnforceOverrides, overrides

from ..container import MetadorContainer
from ..ih5.manifest import IH5MFRecord
from ..plugin import interface as pg
from ..plugins import plugingroups
from ..schema.core import MetadataSchema
from ..schema.plugins import PluginPkgMeta
from ..util.diff import DirDiff
from ..util.hashsums import DirHashsums, dir_hashsums
from .types import DirValidationErrors


class Packer(ABC, EnforceOverrides):
    """Interface to be implemented by Metador HDF5 packer plugins.

    These plugins is how support for wildly different domain-specific
    use-cases can be added to Metador in a opt-in and loosely-coupled way.

    Users can install only the packer plugins they need for their use-cases,
    and such plugins can be easily developed independently from the rest
    of the Metador tooling, as long as this interface is respected.

    Carefully read the documentation for the required attributes and methods
    and implement them for your use-case in a subclass.
    See `metador_core.packer.example.GenericPacker` for an example plugin.

    Requirements for well-behaved packers:

    1. No closing of the container:
    The packer gets a writable record and is only responsible for performing
    the neccessary additions, deletions and modifications. It is not allowed
    to `close()` the container.

    2. No access to data in the container:
    Data in the container MUST NOT be read or be relied on for doing an update,
    as the nodes could be dummy stubs. One MAY rely on existence or absence of
    Groups, Datasets, Attributes and Metadata in the container (e.g. `in` or `keys`).

    3. Source directory is read-only:
    Files or directories inside of `data_dir` MUST NOT be created, deleted or
    modified by this method.

    4. Exceptional termination:
    In case that packing must be aborted, and exception MUST be raised.
    If the exception happened due to invalid data or metadata, it MUST be
    an DirValidationError object like in the other methods above, helping to find
    and fix the problem. Otherwise, a different appropriate exception may be used.

    5. Semantic correctness:
    Packing a directory into a fresh container and updating an existing container
    MUST lead to the same observable result.

    If you cannot guarantee this in full generality, do not implement `update`.
    In that case, if a container is updated, it will be cleared and then `pack` is
    called on it, as if it was a fresh container. In this case, there is no space
    advantage gained over a fresh container (but it will keep its UUID).

    6. Semantic versioning of packers:

    A packer MUST be able to update records that were created by this packer
    of the same or an earlier MINOR version.

    More formally, the version MAJOR.MINOR.PATCH
    MUST adhere to the following contract:

    1. increasing MAJOR means a break in backward-compatibility
    for older datasets (i.e. new packer cannot work with old records),

    2. increasing MINOR means a break in forward-compatibility
    for newer datasets (i.e. older packers will not work with newer records),

    3. increasing PATCH does not affect compatibility
    for datasets with the same MAJOR and MINOR version.

    When the packer is updated, the Python package version MUST increase
    in a suitable way. As usual, whenever an earlier number is increased,
    the following numbers are reset to zero.

    This means, the PATCH version should increase for e.g. bugfixes that do
    not change the structure or metadata stored in the dataset,
    MINOR should increase whenever from now on older versions of the packer
    would not be able to produce a valid update for a dataset created with this version,
    but upgrading older existing datasets with this version still works.
    Finally, MAJOR version is increased when all compatibility guarantees are off
    and the resulting container cannot be migrated or updated automatically.

    You SHOULD provide tooling to migrate datasets between major versions.
    """

    Plugin: PackerPlugin

    @classmethod
    @abstractmethod
    def check_dir(cls, data_dir: Path) -> DirValidationErrors:
        """Check whether the given directory is suitable for packing with this plugin.

        This method will be called before `pack` or `update` and MUST detect
        all problems (such as missing or invalid data or metadata) that can be
        expected to be fixed by the user in preparation for the packing.

        More specifically, it MUST cover all metadata that is to be provided directly by
        the user (i.e. is not inferred or extracted from generated data) for the purpose
        of packing and SHOULD try to cover as many problems with data and metadata as
        possible to avoid failure during the actual packing process.

        Files or directories inside of `data_dir` MUST NOT be created,
        deleted or modified by this method.

        Args:
            data_dir: Directory containing all the data to be packed.

        Returns:
            DirValidationError initialized with a dict mapping file paths
            (relative to `dir`) to lists of detected errors.

            The errors must be either a string (containing a human-readable summary of all
            problems with that file), or another dict with more granular error messages,
            in case that the file is e.g. a JSON-compatible file subject to validation
            with JSON Schemas.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def update(cls, mc: MetadorContainer, data_dir: Path, diff: DirDiff):
        """Update a MetadorContainer with changes done to the data source directory.

        The `container` is assumed to be writable, and is either empty
        or was previously packed by a compatible version of the packer.

        The `data_dir` is assumed to be suitable (according to `check_dir`).

        The `diff` structure contains information about changed paths.

        If not implemented, updates will be created by clearing the provided
        container and using `pack` on it.

        Args:
            mc: Metador IH5 record to pack the data into or update
            data_dir: Directory containing all the data to be packed
            diff: Diff tree of dirs and files in data_dir compared to a previous state
        """
        # default fallback implementation using pack
        for obj in [mc, mc.attrs, mc.meta]:
            for key in obj.keys():
                del obj[key]
        return cls.pack(mc, data_dir)

    @classmethod
    @abstractmethod
    def pack(cls, mc: MetadorContainer, data_dir: Path):
        """Pack a directory into an MetadorContainer.

        The `container` is assumed to be writable and empty.

        The `data_dir` is assumed to be suitable (according to `check_dir`).

        If not implemented, initial packing is done using `update`
        with an empty container and a diff containing all the files.

        Args:
            mc: Metador IH5 record to pack the data into or update
            data_dir: Directory containing all the data to be packed
        """
        # default fallback implementation using update
        return cls.update(mc, data_dir, DirDiff.compare({}, dir_hashsums(data_dir)))


class PackerInfo(MetadataSchema):
    """Schema for info about the packer that was used to create a container."""

    class Plugin:
        name = "core.packerinfo"
        version = (0, 1, 0)

    packer: PGPacker.PluginRef
    """Packer plugin used to pack the container."""

    pkg: PluginPkgMeta
    """Python package that provides the packer plugin."""

    source_dir: DirHashsums = {}
    """Directory skeleton with hashsums of files at the time of packing."""

    @classmethod
    def for_packer(cls, packer_name: str, packer_version=None) -> PackerInfo:
        from ..plugins import packers

        p_ref = packers.resolve(packer_name, packer_version)
        return PackerInfo(
            packer=p_ref,
            pkg=packers.provider(p_ref),
        )


class Unclosable(wrapt.ObjectProxy):
    """Wrapper to prevent packers from closing/completing a container file."""

    _self_MSG = "Packers must not finalize the container!"

    def close(self):
        raise UnsupportedOperation(self._self_MSG)

    # specific for IH5Record subtypes:

    def discard_patch(self):
        raise UnsupportedOperation(self._self_MSG)

    def commit_patch(self):
        raise UnsupportedOperation(self._self_MSG)


PACKER_GROUP_NAME = "packer"


class PackerPlugin(pg.PluginBase):
    ...


class PGPacker(pg.PluginGroup[Packer]):
    """Packer plugin group interface."""

    class Plugin:
        name = PACKER_GROUP_NAME
        version = (0, 1, 0)
        plugin_class = Packer
        plugin_info_class = PackerPlugin
        requires = [
            plugingroups.PluginRef(name="schema", version=(0, 1, 0)),
            plugingroups.PluginRef(name="harvester", version=(0, 1, 0)),
        ]

    _PACKER_INFO_NAME = PackerInfo.Plugin.name

    @overrides
    def check_plugin(self, ep_name: str, plugin: Type[Packer]):
        pg.util.check_implements_method(ep_name, plugin, Packer.check_dir)
        missing_pack = pg.util.implements_method(plugin, Packer.pack)
        missing_update = pg.util.implements_method(plugin, Packer.update)
        if missing_pack and missing_update:
            raise TypeError(f"{ep_name}: Neither pack nor update are implemented!")

    # ----

    def _prepare(self, pname: str, srcdir: Path) -> Tuple[Type[Packer], DirHashsums]:
        """Return packer class and hashsums of given directory.

        Raises an exception if packer is not found or `packer.check_dir` fails.
        """
        packer = self[pname]
        if errs := packer.check_dir(srcdir):
            raise errs
        return (packer, dir_hashsums(srcdir))

    def pack(
        self, packer_name: str, data_dir: Path, target: Path, h5like_cls: Callable
    ):
        """Pack a directory into a container using an installed packer.

        `packer_name` must be an installed packer plugin.

        `data_dir` must be an existing directory suitable for the packer.

        `target` must be a non-existing path and will be passed into `h5like_cls` as-is.

        `h5like_cls` must be a class compatible with MetadorContainer.

        In case an exception happens during packing, notice that no cleanup is done.

        The user is responsible for removing inconsistent files that were created.

        Args:
            packer_name: installed packer plugin name
            data_dir: data source directory
            target: target path for resulting container
            h5like_cls: class to use for creating the container
        """
        packer, hashsums = self._prepare(packer_name, data_dir)
        # use skel_only to enforce stub-compatibility of packer
        container = MetadorContainer(h5like_cls(target, "x")).restrict(skel_only=True)
        packer.pack(Unclosable(container), data_dir)
        self._finalize(packer_name, hashsums, container)

    def update(self, packer_name: str, data_dir: Path, target: Path, h5like_cls):
        """Update a container from its source directory using an installed packer.

        Like `pack`, but the `target` must be a container which can be opened
        with the `h5like_cls` and was packed by a compatible packer.

        In case an exception happens during packing, notice that no cleanup is done
        and if the container has been written to, the changes persist.

        The user is responsible for removing inconsistent files that were created
        and ensuring that the previous state can be restored, e.g. from a backup.
        """
        packer, hashsums = self._prepare(packer_name, data_dir)
        # use skel_only to enforce stub-compatibility of packer
        container = MetadorContainer(h5like_cls(target, "r+")).restrict(skel_only=True)

        # check compatibility
        pinfo = container.meta.get(self._PACKER_INFO_NAME)
        if not pinfo:
            msg = f"Container does not have {self._PACKER_INFO_NAME} metadata!"
        curr_ref = self.resolve(packer_name)
        if not curr_ref.supports(pinfo.packer):
            msg = f"{curr_ref} (installed) does not support {pinfo.packer} (container)!"
            raise ValueError(msg)

        diff = DirDiff.compare(pinfo.source_dir, hashsums)
        packer.update(Unclosable(container), data_dir, diff)
        self._finalize(packer_name, hashsums, container)

    def _finalize(self, pname: str, hsums: DirHashsums, cont: MetadorContainer):
        """Set or update packer info in container and close it."""
        if self._PACKER_INFO_NAME in cont.meta:
            del cont.meta[self._PACKER_INFO_NAME]

        pinfo = PackerInfo.for_packer(pname)
        pinfo.source_dir = hsums
        cont.meta[self._PACKER_INFO_NAME] = pinfo

        if isinstance(cont, IH5MFRecord):
            # when using IH5MFRecord,
            # we want the packerinfo in the manifest, so tooling can decide
            # if a container can be updated without having it available.
            # (default manifest already has enough info for creating stubs)
            cont.manifest.manifest_exts[self.name] = pinfo.dict()

        cont.close()



"""Overlay wrappers to access a virtual record consisting of a base container + patches.

The wrappers take care of dispatching requests to records,
groups and attributes to the correct path.
"""
from __future__ import annotations

import re
from dataclasses import dataclass
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

import h5py
import numpy as np

from ..container.protocols import H5DatasetLike

if TYPE_CHECKING:
    from ..container.protocols import H5GroupLike
    from .record import IH5Record
else:
    IH5Record = Any


# dataset value marking a deleted group, dataset or attribute
DEL_VALUE = np.void(b"\x7f")  # ASCII DELETE

T = TypeVar("T")


def _is_del_mark(val) -> bool:
    return isinstance(val, np.void) and val.tobytes() == DEL_VALUE.tobytes()


def _node_is_del_mark(node) -> bool:
    """Return whether node is marking a deleted group/dataset/attribute value."""
    val = node[()] if isinstance(node, h5py.Dataset) else node
    return _is_del_mark(val)


# attribute key marking group substitution (instead of pass-through default for groups)
# attribute value does not matter, but should be np.Empty(None)
# if present, all children and attributes are interpreted as normal nodes, not as a patch
SUBST_KEY = "\x1a"  # ASCII SUBSTITUTE


def _node_is_virtual(node) -> bool:
    """Virtual node (i.e. transparent and only carrier for child nodes and attributes)."""
    return isinstance(node, h5py.Group) and SUBST_KEY not in node.attrs


@dataclass(frozen=True)
class IH5Node:
    """An overlay node wraps a group, dataset or attribute manager.

    It takes care of finding the correct container to look for the data
    and helps with patching data in a new patch container.

    It essentially lifts the interface of h5py from a single file to an IH5 record
    that may consist of a base container file and a number of patch containers.
    """

    _record: IH5Record
    """Record this node belongs to (needed to access the actual data)."""

    _gpath: str
    """Path in record that this node represents (absolute wrt. root of record)."""

    _cidx: int
    """Left boundary index for lookups in order of loaded containers, i.e.
    this node will not consider containers with smaller index than that.
    """

    def __post_init__(self):
        """Instantiate an overlay node."""
        if not self._gpath or self._gpath[0] != "/":
            raise ValueError("Path must be absolute!")
        if self._cidx < 0:
            raise ValueError("Creation index must be non-negative!")

    @property
    def _files(self) -> List[h5py.File]:
        return self._record.__files__

    def __hash__(self):
        """Hash an overlay node.

        Two nodes are equivalent if they are linked to the same
        open record and address the same entity.
        """
        return hash((id(self._record), self._gpath, self._cidx))

    def __bool__(self) -> bool:
        return bool(self._files) and all(map(bool, self._files))

    @property
    def _last_idx(self):
        """Index of the latest container."""
        return len(self._files) - 1

    @property
    def _is_read_only(self) -> bool:
        """Return true if the newest container is read-only and nothing can be written."""
        return not self._record._has_writable

    def _guard_open(self):
        """Check that the record is open (if it was closed, the files are gone)."""
        if not self:
            raise ValueError("Record is not open or accessible!")

    def _guard_read_only(self):
        if self._is_read_only:
            raise ValueError("Create a patch in order to change the container!")

    def _guard_value(self, data):
        if _is_del_mark(data):
            raise ValueError(f"Value '{data}' is forbidden, cannot assign!")
        if isinstance(data, IH5Node):
            raise ValueError("Hard links are not supported, cannot assign!")
        if isinstance(data, h5py.SoftLink) or isinstance(data, h5py.ExternalLink):
            raise ValueError("SymLink and ExternalLink not supported, cannot assign!")

    @classmethod
    def _latest_idx(cls, files, path: str) -> Optional[int]:
        """Return index of newest file where the group/dataset was overwritten/created.

        Returns None if not found or most recent value is a deletion mark.
        """
        idx = None
        for i in reversed(range(len(files))):
            if _node_is_del_mark(files[i][path]):
                return None
            elif _node_is_virtual(files[i][path]):
                idx = i
            else:
                return i  # some patch overrides the group
        return idx

    # path transformations

    def _parent_path(self) -> str:
        """Return path of the parent node (the root is its own parent)."""
        if self._gpath == "/":
            return "/"
        segs = self._gpath.split("/")[:-1]
        return "/" if segs == [""] else "/".join(segs)

    def _rel_path(self, path: str) -> str:
        """Return relative path based on node location, if passed path is absolute.

        If relative, returns the path back unchanged.
        """
        if path[0] != "/":
            return path
        if path.find(self._gpath) != 0:
            raise RuntimeError("Invalid usage, cannot strip non-matching prefix!")
        start_idx = len(self._gpath) + int(self._gpath != "/")
        return path[start_idx:]

    def _abs_path(self, path: str) -> str:
        """Return absolute path based on node location, if given path is relative.

        If absolute, returns the path back unchanged.
        """
        pref = self._gpath if self._gpath != "/" else ""
        return path if path and path[0] == "/" else f"{pref}/{path}"

    def _inspect_path(self, path):  # pragma: no cover
        """Print the path node of all containers where the path is contained in."""
        print(f"Path {path}:")
        for j in range(len(self._files)):
            if path in self._files[j]:
                node = self._files[j][path]
                print(f"  idx={j}: {type(node).__name__}")
                if isinstance(node, h5py.Dataset):
                    print("    ", node[()])


class IH5InnerNode(IH5Node):
    """Common functionality for Group and AttributeManager.

    Will grant either access to child records/subgroups,
    or to the attributes attached to the group/dataset at a path in a record.
    """

    @property
    def _is_attrs(self) -> bool:
        return self.__is_attrs__

    def __init__(
        self,
        record: IH5Record,
        gpath: str,
        creation_idx: int,
        attrs: bool = False,
    ):
        """See `IH5Node` constructor.

        This variant represents an "overlay container", of which there are two types -
        a group (h5py.Group) and a set of attributes (h5py.AttributeManager).

        This class takes care of both (in order to avoid lots of code duplication),
        distinguishing them through the additional `attrs` flag.
        """
        super().__init__(record, gpath, creation_idx)
        # if attrs set, represents AttributeManager, otherwise its a group
        self.__is_attrs__: bool = attrs

    def _guard_key(self, key: str):
        """Check a key used with bracket accessor notation.

        (e.g. used for `__getitem__, __setitem__, __delitem__`)
        """
        if key == "":
            raise ValueError("Invalid empty path!")
        if key.find("@") >= 0:  # used as attribute separator in the skeleton! TODO
            raise ValueError(f"Invalid symbol '@' in key: '{key}'!")
        if re.match(r"^[!-~]+$", key) is None:
            raise ValueError("Invalid key: Only printable ASCII is allowed!")
        if self._is_attrs and (key.find("/") >= 0 or key == SUBST_KEY):
            raise ValueError(f"Invalid attribute key: '{key}'!")

    def _get_child_raw(self, key: str, cidx: int) -> Any:
        """Return given child (dataset, group, attribute) from given container."""
        if self._is_attrs:
            return self._files[cidx][self._gpath].attrs[key]
        else:
            return self._files[cidx][self._abs_path(key)]

    def _get_child(self, key: str, cidx: int) -> Any:
        """Like _get_child_raw, but wraps the result with an overlay class if needed."""
        val = self._get_child_raw(key, cidx)
        path = self._abs_path(key)
        if isinstance(val, h5py.Group):
            return IH5Group(self._record, path, cidx)
        elif isinstance(val, h5py.Dataset):
            return IH5Dataset(self._record, path, cidx)
        else:
            return val

    def _children(self) -> Dict[str, int]:
        """Return dict mapping from a child name to the most recent overriding patch idx.

        For datasets, dereferencing the child path in that container will give the data.
        For groups, the returned number is to be treated as the lower bound, i.e.
        the child creation_idx to recursively get the descendents.
        """
        self._guard_open()

        children: Dict[str, int] = {}
        is_virtual: Dict[str, bool] = {}
        for i in reversed(range(self._cidx, len(self._files))):
            if self._gpath not in self._files[i]:
                continue

            obj = self._files[i][self._gpath]
            if self._is_attrs:
                obj = obj.attrs
            assert isinstance(obj, (h5py.Group, h5py.AttributeManager))

            # keep most recent version of child node / attribute
            for k in obj.keys():
                if k not in children:
                    is_virtual[k] = _node_is_virtual(self._get_child_raw(k, i))
                    children[k] = i
                elif is_virtual[k]:  # .. and k in children!
                    # decrease lower bound
                    children[k] = min(children[k], i)

        # return resulting child nodes / attributes (without the deleted ones)
        # in alphabetical order,
        # in case of attributes, also excludes special SUBST marker attribute
        return {
            k: idx
            for k, idx in sorted(children.items(), key=lambda x: x[0])
            if (not self._is_attrs or k != SUBST_KEY)
            and not _node_is_del_mark(self._get_child_raw(k, idx))
        }

    def _get_children(self) -> List[Any]:
        """Get alphabetically ordered list of child nodes."""
        return [
            self._get_child(self._abs_path(k), idx)
            for k, idx in self._children().items()
        ]

    def _node_seq(self, path: str) -> List[IH5Node]:
        """Return node sequence (one node per path prefix) to given path.

        Returns:
            Sequence starting with the current node (if path is relative)
            or the root node (if absolute) followed by all successive
            children along the requested path that exist.
        """
        curr: IH5InnerNode = IH5Group(self._record) if path[0] == "/" else self

        ret: List[IH5Node] = [curr]
        if path == "/" or path == ".":  # special case
            return ret

        # access entity through child group sequence
        segs = path.strip("/").split("/")
        nxt_cidx = 0
        for i in range(len(segs)):
            seg, is_last_seg = segs[i], i == len(segs) - 1
            # find most recent container with that child
            nxt_cidx = curr._children().get(seg, -1)
            if nxt_cidx == -1:
                return ret  # not found -> return current prefix
            curr = curr._get_child(seg, nxt_cidx)  # proceed to child
            ret.append(curr)
            # catch invalid access, e.g. /foo is record, user accesses /foo/bar:
            if not is_last_seg and isinstance(curr, IH5Dataset):
                raise ValueError(f"Cannot access path inside a value: {curr._gpath}")
        # return path index sequence
        return ret

    def _find(self, key: str) -> Optional[int]:
        """Return index of container holding that key (attribute or path), if any.

        Args:
            key: nonempty string (attribute, or relative/absolute path)

        Returns:
            Index >= 0 of most recent container patching that path if found, else None.
        """
        if self._is_attrs:  # access an attribute by key (always "relative")
            return self._children().get(key, None)
        # access a path (absolute or relative)
        nodes = self._node_seq(key)
        return nodes[-1]._cidx if nodes[-1]._gpath == self._abs_path(key) else None

    # h5py-like interface

    def get(self, key: str, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key: str):
        self._guard_open()
        self._guard_key(key)
        found_cidx = self._find(key)
        if found_cidx is None:
            raise KeyError(key)
        return self._get_child(key, found_cidx)

    def _expect_real_item_idx(self, key: str) -> int:
        found_cidx = self._find(key)
        if found_cidx is None or _node_is_del_mark(self._get_child(key, found_cidx)):
            raise KeyError(f"Cannot delete '{key}', it does not exist!")
        return found_cidx

    def __contains__(self, key: str):
        self._guard_key(key)
        return self._find(key) is not None

    def __iter__(self):
        return iter(self._children().keys())

    def __len__(self):
        return len(self.keys())

    def keys(self):
        return self._children().keys()

    def _dict(self):
        return {k: self._get_child(k, idx) for k, idx in self._children().items()}

    def values(self):
        return self._dict().values()

    def items(self):
        return self._dict().items()


class IH5Dataset(IH5Node):
    """`IH5Node` representing a `h5py.Dataset`, i.e. a leaf of the tree."""

    def __init__(self, files, gpath, creation_idx):
        super().__init__(files, gpath, creation_idx)

    def copy_into_patch(self):
        """Copy the most recent value at this path into the current patch.

        This is useful e.g. for editing inside a complex value, such as an array.
        """
        self._guard_open()
        self._guard_read_only()
        if self._cidx == self._last_idx:
            raise ValueError("Cannot copy, this node is already from latest patch!")
        # copy value from older container to current patch
        self._files[-1][self._gpath] = self[()]

    # h5py-like interface
    @property
    def name(self) -> str:
        return self._gpath

    @property
    def file(self) -> IH5Record:
        return self._record

    @property
    def parent(self) -> IH5Group:
        return self._record[self._parent_path()]

    @property
    def attrs(self) -> IH5AttributeManager:
        self._guard_open()
        return IH5AttributeManager(self._record, self._gpath, self._cidx)

    # this one is also needed to work with H5DatasetLike
    @property
    def ndim(self) -> int:
        return self._files[self._cidx][self._gpath].ndim  # type: ignore

    # for a dataset, instead of paths the numpy data is indexed. at this level
    # the patching mechanism ends, so it's just passing through to h5py

    def __getitem__(self, key):
        # just pass through dataset indexing to underlying dataset
        self._guard_open()
        return self._files[self._cidx][self._gpath][key]  # type: ignore

    def __setitem__(self, key, val):
        self._guard_open()
        self._guard_read_only()
        if self._cidx != self._last_idx:
            raise ValueError(f"Cannot set '{key}', node is not from the latest patch!")
        # if we're in the latest patch, allow writing as usual (pass through)
        self._files[-1][self._gpath][key] = val  # type: ignore


class IH5AttributeManager(IH5InnerNode):
    """`IH5Node` representing an `h5py.AttributeManager`."""

    def __init__(self, files, gpath, creation_idx):
        super().__init__(files, gpath, creation_idx, True)

    def __setitem__(self, key: str, val):
        self._guard_open()
        self._guard_read_only()
        self._guard_key(key)
        self._guard_value(val)

        # if path does not exist in current patch, just create "virtual node"
        if self._gpath not in self._files[-1]:
            self._files[-1].create_group(self._gpath)
        # deletion marker at `key` (if set) is overwritten automatically here
        # so no need to worry about removing it before assigning `val`
        self._files[-1][self._gpath].attrs[key] = val

    def __delitem__(self, key: str):
        self._guard_open()
        self._guard_read_only()
        self._guard_key(key)
        # remove the entity if it is found in newest container,
        # mark the path as deleted if doing a patch and not working on base container
        if self._expect_real_item_idx(key) == self._last_idx:
            del self._files[-1][self._gpath].attrs[key]
        if len(self._files) > 1:  # is a patch?
            if self._gpath not in self._files[-1]:  # no node at path in latest?
                self._files[-1].create_group(self._gpath)  # create "virtual" node
            self._files[-1][self._gpath].attrs[key] = DEL_VALUE  # mark deleted


class IH5Group(IH5InnerNode):
    """`IH5Node` representing a `h5py.Group`."""

    def _require_node(self, name: str, node_type: Type[T]) -> Optional[T]:
        # helper for require_{group|dataset}
        grp = self.get(name)
        if isinstance(grp, node_type):
            return grp
        if grp is not None:
            msg = f"Incompatible object ({type(grp).__name__}) already exists"
            raise TypeError(msg)
        return None

    def __init__(self, record, gpath: str = "/", creation_idx: Optional[int] = None):
        if gpath == "/":
            creation_idx = 0
        if creation_idx is None:
            raise ValueError("Need creation_idx for path != '/'!")
        super().__init__(record, gpath, creation_idx, False)

    def _create_virtual(self, path: str) -> bool:
        nodes = self._node_seq(path)
        path = self._abs_path(path)
        if (
            nodes[-1]._gpath == path
            and nodes[-1]._cidx == self._last_idx
            and not _node_is_del_mark(nodes[-1])
        ):
            return False  # something at that path in most recent container exists

        # most recent entity is a deletion marker or not existing?
        if nodes[-1]._gpath != path or _node_is_del_mark(nodes[-1]):
            suf_segs = nodes[-1]._rel_path(path).split("/")
            # create "overwrite" group in most recent patch...
            self.create_group(f"{nodes[-1]._gpath}/{suf_segs[0]}")
            # ... and create (nested) virtual group node(s), if needed
            if len(suf_segs) > 1:
                self._files[-1].create_group(path)

        return True

    # h5py-like interface

    def __setitem__(self, path: str, value):
        return self.create_dataset(path, data=value)

    def __delitem__(self, key: str):
        self._guard_open()
        self._guard_read_only()
        self._guard_key(key)
        self._expect_real_item_idx(key)
        # remove the entity if it is found in newest container,
        # mark the path as deleted if doing a patch and not working on base container
        path = self._abs_path(key)
        if path in self._files[-1]:
            del self._files[-1][path]
        if len(self._files) > 1:  # has patches? mark deleted (instead of real delete)
            self._files[-1][path] = DEL_VALUE

    @property
    def name(self) -> str:
        return self._gpath

    @property
    def file(self):  # -> IH5Record
        return self._record

    @property
    def parent(self) -> IH5Group:
        return self._record[self._parent_path()]

    @property
    def attrs(self) -> IH5AttributeManager:
        self._guard_open()
        return IH5AttributeManager(self._record, self._gpath, self._cidx)

    def create_group(self, name: str) -> IH5Group:
        self._guard_open()
        self._guard_read_only()

        path = self._abs_path(name)
        nodes = self._node_seq(path)
        if not isinstance(nodes[-1], IH5Group):
            raise ValueError(f"Cannot create group, {nodes[-1]._gpath} is a dataset!")
        if nodes[-1]._gpath == path:
            raise ValueError("Cannot create group, it already exists!")

        # remove "deleted" marker, if set at current path in current patch container
        if path in self._files[-1] and _node_is_del_mark(self._files[-1][path]):
            del self._files[-1][path]
        # create group (or fail if something else exists there already)
        self._files[-1].create_group(path)
        # if this is a patch: mark as non-virtual, i.e. "overwrite" with empty group
        # because the intent here is to "create", not update something.
        if len(self._files) > 1:
            self._files[-1][path].attrs[SUBST_KEY] = h5py.Empty(None)

        return IH5Group(self._record, path, self._last_idx)

    def create_dataset(
        self, path: str, shape=None, dtype=None, data=None, **kwargs
    ) -> IH5Dataset:
        self._guard_open()
        self._guard_read_only()
        self._guard_key(path)
        self._guard_value(data)

        if unknown_kwargs := set(kwargs.keys()) - {"compression", "compression_opts"}:
            raise ValueError(f"Unkown kwargs: {unknown_kwargs}")

        path = self._abs_path(path)
        fidx = self._find(path)
        if fidx is not None:
            prev_val = self._get_child(path, fidx)
            if isinstance(prev_val, (IH5Group, IH5Dataset)):
                raise ValueError("Path exists, in order to replace - delete first!")

        if path in self._files[-1] and _node_is_del_mark(
            self._get_child_raw(path, self._last_idx)
        ):
            # remove deletion marker in latest patch, if set
            del self._files[-1][path]
        elif path not in self._files[-1]:
            # create path and overwrite-group in latest patch
            self._create_virtual(path)
            assert path in self._files[-1]
            del self._files[-1][path]

        self._files[-1].create_dataset(  # actually create it, finally
            path, shape=shape, dtype=dtype, data=data, **kwargs
        )
        return IH5Dataset(self._record, path, self._last_idx)

    def require_group(self, name: str) -> IH5Group:
        if (n := self._require_node(name, IH5Group)) is not None:
            return n  # existing group
        return self.create_group(name)

    def require_dataset(self, name: str, *args, **kwds) -> IH5Dataset:
        if (n := self._require_node(name, IH5Dataset)) is not None:
            # TODO: check dimensions etc, copy into patch if it fits
            return n
        return self.create_dataset(name, *args, **kwds)

    def copy(self, source: CopySource, dest: CopyDest, **kwargs):
        src_node = self[source] if isinstance(source, str) else source
        name: str = kwargs.pop("name", src_node.name.split("/")[-1])
        dst_name: str
        if isinstance(dest, str):
            # if dest is a path, ignore inferred/passed name
            segs = self._abs_path(dest).split("/")
            dst_group = self.require_group("/".join(segs[:-1]) or "/")
            dst_name = segs[-1]
        else:
            # given dest is a group node, use inferred/passed name

            dst_group = dest if dest.name != "/" else dest["/"]  # *
            # * ugly workaround for treating files as groups in the copy method

            dst_name = name
        return h5_copy_from_to(src_node, cast(Any, dst_group), dst_name, **kwargs)

    def move(self, source: str, dest: str):
        self.copy(source, dest)
        del self[source]

    def visititems(self, func: Callable[[str, object], Optional[Any]]) -> Any:
        self._guard_open()
        stack = list(reversed(self._get_children()))
        while stack:
            curr = stack.pop()
            val = func(self._rel_path(curr._gpath), curr)
            if val is not None:
                return val
            if isinstance(curr, IH5Group):
                stack += reversed(curr._get_children())

    def visit(self, func: Callable[[str], Optional[Any]]) -> Any:
        return self.visititems(lambda x, _: func(x))


CopySource = Union[str, IH5Group, IH5Dataset, h5py.Group, h5py.Dataset]
CopyDest = Union[str, IH5Group, h5py.Group]


# ----
# Helpers for IH5 / H5 interop (its all h5py at the bottom anyway, so its easy)


class H5Type(str, Enum):
    """Type of an entity in a HDF5-like container.

    We list only those we care about, ignoring various
    link types etc.

    This will be used in wrappers around HDF5-like objects
    instead of using isinstance/subclass checks to implement
    duck-typing based decorator functionality that can
    work with (at least) raw HDF5, IH5 and IH5+Manifest.
    """

    group = "group"  # possibly nested, dict-like
    dataset = "dataset"  # = wrapped, indexable data
    attribute_set = "attribute-set"  # = not further nested, dict-like
    attribute = "attribute"  # = unwrapped data

    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.value}"


def h5_copy_from_to(
    source_node: Union[H5DatasetLike, H5GroupLike],
    target_group: H5GroupLike,
    target_path: str,
    **kwargs,
):
    """Copy a dataset or group from one container to a fresh location.

    This works also between HDF5 and IH5.

    Source node must be group or dataset object.
    Target node must be an existing group object.
    Target path must be fresh path relative to target node.
    """
    without_attrs: bool = kwargs.pop("without_attrs", False)
    shallow: bool = kwargs.pop("shallow", False)
    for arg in ["expand_soft", "expand_external", "expand_refs"]:
        if not kwargs.pop(arg, True):
            raise ValueError("IH5 does not support keeping references!")
    if kwargs:
        raise ValueError(f"Unknown keyword arguments: {kwargs}")

    if not target_path or target_path[0] == "/":
        raise ValueError("Target path must be non-empty and relative!")
    if target_path in target_group:
        raise ValueError(f"Target path {target_path} already exists in target group!")

    def copy_attrs(src_node, trg_node):
        if not without_attrs:
            trg_atrs = trg_node.attrs
            for k, v in src_node.attrs.items():
                trg_atrs[k] = v

    if isinstance(source_node, H5DatasetLike):
        node = target_group.create_dataset(target_path, data=source_node[()])
        copy_attrs(source_node, node)  # copy dataset attributes
    else:
        trg_root = target_group.create_group(target_path)
        copy_attrs(source_node, trg_root)  # copy source node attributes

        def copy_children(name, src_child):
            # name is relative to source root -> can use it
            if isinstance(src_child, H5DatasetLike):
                trg_root[name] = src_child[()]
            else:  # must be grouplike
                trg_root.create_group(name)
            copy_attrs(src_child, trg_root[name])

        if shallow:  # only immediate children
            for name, src_child in source_node.items():
                copy_children(name, src_child)
        else:  # recursive copy
            source_node.visititems(copy_children)



"""Sidecar JSON file storing a skeleton to create stubs and patch containers."""
from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from uuid import UUID, uuid1

from pydantic import BaseModel

from ..schema.types import QualHashsumStr
from ..util.hashsums import qualified_hashsum
from .record import IH5Record, IH5UserBlock, hashsum_file
from .skeleton import IH5Skeleton, init_stub_base


class IH5Manifest(BaseModel):
    """A metadata sidecar file for a collection of IH5 record files.

    It contains the skeleton of the container, in order to be able to create a
    patch for an IH5Record the data is not locally available for (but manifest is).
    """

    # uuid for the manifest file itself (so the filename does not matter)
    manifest_uuid: UUID

    user_block: IH5UserBlock  # copy of user block (without the manifest extension part)

    skeleton: IH5Skeleton  # computed with IH5Skeleton, used to create a stub

    manifest_exts: Dict[str, Any]  # Arbitrary extensions, similar to IH5UserBlock

    @classmethod
    def from_userblock(cls, ub: IH5UserBlock, skeleton=None, exts=None) -> IH5Manifest:
        """Create a manifest file based on a user block."""
        skeleton = skeleton or {}
        exts = exts or {}
        ub_copy = ub.copy()
        # only keep other extensions (otherwise its circular)
        ub_copy.ub_exts = {
            k: v for k, v in ub.ub_exts.items() if k != IH5UBExtManifest.ext_name()
        }
        return cls(
            manifest_uuid=uuid1(),
            user_block=ub_copy,
            skeleton=skeleton,
            manifest_exts=exts,
        )

    def __bytes__(self) -> bytes:
        """Serialize to JSON and return UTF-8 encoded bytes to be written in a file."""
        # add a newline, as otherwise behaviour with text editors will be confusing
        # (e.g. vim automatically adds a trailing newline that it hides)
        # https://stackoverflow.com/questions/729692/why-should-text-files-end-with-a-newline
        return (self.json(indent=2) + "\n").encode(encoding="utf-8")

    def save(self, path: Path):
        """Save manifest (as returned by bytes()) into a file."""
        with open(path, "wb") as f:
            f.write(bytes(self))
            f.flush()


class IH5UBExtManifest(BaseModel):
    """IH5 user block extension for stub and manifest support."""

    is_stub_container: bool
    """True if file has the structure of another container, without actual data inside."""

    manifest_uuid: UUID
    """UUID of the manifest file that belongs to this IH5 file."""

    manifest_hashsum: QualHashsumStr
    """Hashsum of the manifest file that belongs to this IH5 file."""

    @classmethod
    def ext_name(cls) -> str:
        """Name of user block extension section for stub and manifest info."""
        return "ih5mf_v01"

    @classmethod
    def get(cls, ub: IH5UserBlock) -> Optional[IH5UBExtManifest]:
        """Parse extension metadata from userblock, if it is available."""
        if cls.ext_name() not in ub.ub_exts:
            return None
        return cls.parse_obj(ub.ub_exts[cls.ext_name()])

    def update(self, ub: IH5UserBlock):
        """Create or overwrite extension metadata in given userblock."""
        ub.ub_exts[self.ext_name()] = self.dict()


class IH5MFRecord(IH5Record):
    """IH5Record extended by a manifest file.

    The manifest file is a sidcar JSON file that contains enough information to support
    the creation of a stub container and patching a dataset without having the actual
    container locally available.

    In a chain of container files, only the base container may be a stub.
    All files without the manifest extension in the userblock are considered not stubs.

    An IH5MFRecord is a valid IH5Record (the manifest file then is simply ignored).
    Also, it is possible to open an IH5Record as IH5MFRecord and turn it into a
    valid IH5MFRecord by committing a patch (this will create the missing manifest).

    In addition to the ability to create stubs, the manifest file can be used to carry
    information that should be attached to a container, but is too large or inappropriate
    for storage in the userblock (e.g. should be available separately).

    The manifest should store information that applies semantically to the whole fileset
    at the current patch level, it MUST NOT be required to have manifest files for each
    ih5 patch. Additional information stored in the manifest is inherited to the
    manifest of successive patches until overridden.

    The main use case of this extension is to be used by automated packers
    from suitably prepared source directories, to be uploaded to remote locations,
    and for these packers being able to create patches for the record without
    access to all the data containers (based only on the most recent manifest file).
    """

    MANIFEST_EXT: str = "mf.json"

    _manifest: Optional[IH5Manifest] = None
    """Manifest of newest loaded container file (only None for new uncommited records)."""

    @property
    def manifest(self) -> IH5Manifest:
        """Return loaded manifest object of latest committed record patch."""
        if self._manifest is None:  # should only happen with fresh create()d records
            raise ValueError("No manifest exists yet! Did you forget to commit?")
        return self._manifest

    def _fresh_manifest(self) -> IH5Manifest:
        """Return new manifest based on current state of the record."""
        ub = self._ublock(-1)
        skel = IH5Skeleton.for_record(self)
        return IH5Manifest.from_userblock(ub, skeleton=skel, exts={})

    @classmethod
    def _manifest_filepath(cls, record: str) -> Path:
        """Return canonical filename of manifest based on path of a container file."""
        return Path(f"{str(record)}{cls.MANIFEST_EXT}")

    # Override to also load and check latest manifest
    @classmethod
    def _open(cls, paths: List[Path], **kwargs):
        manifest_file: Optional[Path] = kwargs.pop("manifest_file", None)
        ret: IH5MFRecord = super()._open(paths, **kwargs)

        # if not given explicitly, infer correct manifest filename
        # based on logically latest container (they are sorted after parent init)
        if manifest_file is None:
            manifest_file = cls._manifest_filepath(ret._files[-1].filename)

        # for latest container, check linked manifest (if any) against given/inferred one
        ub = ret._ublock(-1)
        ubext = IH5UBExtManifest.get(ub)
        if ubext is not None:
            if not manifest_file.is_file():
                msg = f"Manifest file {manifest_file} does not exist, cannot open!"
                raise ValueError(f"{ret._files[-1].filename}: {msg}")

            chksum = hashsum_file(manifest_file)
            if ubext.manifest_hashsum != chksum:
                msg = "Manifest has been modified, unexpected hashsum!"
                raise ValueError(f"{ret._files[-1].filename}: {msg}")

            ret._manifest = IH5Manifest.parse_file(manifest_file)
            # NOTE: as long as we enforce checksum of manifest, this failure can't happen:
            # if ubext.manifest_uuid != self._manifest.manifest_uuid:
            #     raise ValueError(f"{ub._filename}: Manifest file has wrong UUID!")
        # all looks good
        return ret

    # Override to also check user block extension
    def _check_ublock(
        self,
        filename: Union[str, Path],
        ub: IH5UserBlock,
        prev: Optional[IH5UserBlock] = None,
        check_hashsum: bool = True,
    ):
        super()._check_ublock(filename, ub, prev, check_hashsum)
        # Try getting manifest info in the userblock.
        # If it is missing, probably we're opening a "raw" IH5Record or a messed up mix
        ubext = IH5UBExtManifest.get(ub)
        # we only allow to write patches on top of stubs,
        # but not have stubs on top of something else.
        # If something creates a patch that is (marked as) a stub, its a developer error.
        # If the ub ext is missing, then we must assume that it is not a stub.
        assert prev is None or ubext is None or not ubext.is_stub_container

    def _fixes_after_merge(self, file, ub):
        # if a manifest exists for the current dataset,
        # copy its manifest to overwrite the fresh one of the merged container
        # and fix its user block
        if self._manifest is not None:
            # check that new userblock inherited the original linked manifest
            ext = IH5UBExtManifest.get(ub)
            assert ext is not None and ext.manifest_uuid == self.manifest.manifest_uuid
            # overwrite the "fresh" manifest from merge with the original one
            self.manifest.save(self._manifest_filepath(file))

    # Override to prevent merge if a stub is present
    def merge_files(self, target: Path):
        def is_stub(x):
            ext = IH5UBExtManifest.get(x)
            # missing ext -> not a stub (valid stub has ext + is marked as stub)
            return ext is not None and ext.is_stub_container

        if any(map(is_stub, self.ih5_meta)):
            raise ValueError("Cannot merge, files contain a stub!")

        return super().merge_files(target)

    # Override to create skeleton and dir hashsums, write manifest and add to user block
    # Will inherit old manifest extensions, unless overridden by passed argument
    def commit_patch(self, **kwargs) -> None:
        # is_stub == True only if called from create_stub!!! (NOT for the "end-user"!)
        is_stub = kwargs.pop("__is_stub__", False)
        exts = kwargs.pop("manifest_exts", None)

        # create manifest for the new patch
        mf = self._fresh_manifest()
        if self._manifest is not None:  # inherit attached data, if manifest exists
            mf.manifest_exts = self.manifest.manifest_exts
        if exts is not None:  # override, if extensions provided
            mf.manifest_exts = exts

        old_ub = self._ublock(-1)  # keep ref in case anything goes wrong
        # prepare new user block that links to the prospective manifest
        new_ub = old_ub.copy()
        IH5UBExtManifest(
            is_stub_container=is_stub,
            manifest_uuid=mf.manifest_uuid,
            manifest_hashsum=qualified_hashsum(bytes(mf)),
        ).update(new_ub)

        # try writing new container
        self._set_ublock(-1, new_ub)
        try:
            super().commit_patch(**kwargs)
        except ValueError as e:  # some checks failed
            self._set_ublock(-1, old_ub)  # reset current user block
            raise e

        # as everything is fine, finally (over)write manifest here and on disk
        self._manifest = mf
        mf.save(self._manifest_filepath(self._files[-1].filename))

    @classmethod
    def create_stub(
        cls,
        record: Union[Path, str],
        manifest_file: Path,
    ) -> IH5MFRecord:
        """Create a stub base container for patching an existing but unavailable record.

        The stub is based on the user block of a real IH5 record container line
        and the skeleton of the overlay structure (as returned by `IH5Skeleton`),
        which are taken from a provided manifest file.

        Patches created on top of the stub are compatible with the original record
        whose metadata the stub is based on.

        The returned container is read-only and only serves as base for patches.
        """
        manifest = IH5Manifest.parse_file(manifest_file)

        skeleton: IH5Skeleton = manifest.skeleton
        user_block: IH5UserBlock = manifest.user_block.copy()

        # the manifest-stored user block has no manifest extension itself - create new
        # based on passed manifest.
        # mark it as stub in extra metadata now! important to avoid accidents!
        # must pass it in like that, because the container will be auto-commited.
        ubext = IH5UBExtManifest(
            is_stub_container=True,  # <- the ONLY place where this is allowed!
            manifest_uuid=manifest.manifest_uuid,
            manifest_hashsum=hashsum_file(manifest_file),
        )
        ubext.update(user_block)

        # create and finalize the stub (override userblock and create skeleton structure)
        ds = IH5MFRecord._create(Path(record))
        init_stub_base(ds, user_block, skeleton)  # prepares structure and user block
        # commit_patch() completes stub + fixes the hashsum
        ds.commit_patch(__is_stub__=True)
        assert not ds._has_writable

        return ds



"""Re-exports for users of the IH5 variants."""

# this must be separate to avoid ciruclar imports

from .manifest import IH5Manifest, IH5MFRecord  # noqa: F401
from .overlay import IH5AttributeManager, IH5Dataset, IH5Group  # noqa: F401
from .record import IH5Record, IH5UserBlock  # noqa: F401

__all__ = ["IH5Dataset", "IH5Group", "IH5AttributeManager", "IH5Record"]



"""Partial pydantic models.

Partial models replicate another model with all fields made optional.
However, they cannot be part of the inheritance chain of the original
models, because semantically it does not make sense or leads to
a diamond problem.

Therefore they are independent models, unrelated to the original
models, but convertable through methods.

Partial models are implemented as a mixin class to be combined
with the top level base model that you are using in your model
hierarchy, e.g. if you use plain `BaseModel`, you can e.g. define:

```
class MyPartial(DeepPartialModel, BaseModel): ...
```

And use `MyPartial.get_partial` on your models.

If different compatible model instances are merged,
the merge will produce an instance of the left type.

Some theory - partial schemas form a monoid with:
* the empty partial schema as neutral element
* merge of the fields as the binary operation
* associativity follows from associativity of used merge operations
"""

from __future__ import annotations

from functools import reduce
from typing import (
    Any,
    ClassVar,
    Dict,
    ForwardRef,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from pydantic import BaseModel, ValidationError, create_model, validate_model
from pydantic.fields import FieldInfo
from typing_extensions import Annotated

from ..util import is_public_name
from ..util import typing as t


def _is_list_or_set(hint):
    return t.is_list(hint) or t.is_set(hint)


def _check_type_mergeable(hint, *, allow_none: bool = False) -> bool:
    """Check whether a type is mergeable.

    An atomic type is:
    * not None
    * not a List, Set, Union or Optional
    * not a model

    (i.e. usually a primitive value e.g. int/bool/etc.)

    A singular type is:
    * an atomic type, or
    * a model, or
    * a Union of multiple singular types

    A complex type is:
    * a singular type
    * it is a List/Set of a singular type

    A mergeable type is:
    * a complex type, or
    * an Optional complex type

    merge(x,y):
        merge(None, None) = None
        merge(None, x) = merge(x, None) = x
        merge(x: model1, y: model2)
        | model1 < model2 or model2 < model1 = recursive_merge(x, y)
        | otherwise = y
        merge(x: singular, y: singular) = y
        merge(x: list, y: list) = x ++ y
        merge(x: set, y: set) =  x.union(y)
    """
    # print("check ", hint)
    args = t.get_args(hint)

    if _is_list_or_set(hint):  # list or set -> dig deeper
        return all(map(_check_type_mergeable, args))

    # not union, list or set?
    if not t.is_union(hint):
        # will be either primitive or recursively merged -> ok
        return True

    # Union case:
    if not allow_none and t.is_optional(hint):
        return False  # allows none, but should not!

    # If a Union contains a set or list, it must be only combined with None
    is_prim_union = not any(map(_is_list_or_set, args))
    is_opt_set_or_list = len(args) == 2 and t.NoneType in args
    if not (is_prim_union or is_opt_set_or_list):
        return False

    # everything looks fine on this level -> recurse down
    return all(map(_check_type_mergeable, args))


def is_mergeable_type(hint) -> bool:
    """Return whether given type can be deeply merged.

    This imposes some constraints on the shape of valid hints.
    """
    return _check_type_mergeable(hint, allow_none=True)


def val_from_partial(val):
    """Recursively convert back from a partial model if val is one."""
    if isinstance(val, PartialModel):
        return val.from_partial()
    if isinstance(val, list):
        return [val_from_partial(x) for x in val]
    if isinstance(val, set):
        return {val_from_partial(x) for x in val}
    return val


class PartialModel:
    """Base partial metadata model mixin.

    In this variant merging is done by simply overwriting old field values
    with new values (if the new value is not `None`) in a shallow way.

    For more fancy merging, consider `DeepPartialModel`.
    """

    __partial_src__: ClassVar[Type[BaseModel]]
    """Original model class this partial class is based on."""

    __partial_fac__: ClassVar[Type[PartialFactory]]
    """Factory class that created this partial."""

    def from_partial(self):
        """Return a new non-partial model instance (will run full validation).

        Raises ValidationError on failure (e.g. if the partial is missing fields).
        """
        fields = {
            k: val_from_partial(v)
            for k, v in self.__partial_fac__._get_field_vals(self)
        }
        return self.__partial_src__.parse_obj(fields)

    @classmethod
    def to_partial(cls, obj, *, ignore_invalid: bool = False):
        """Transform `obj` into a new instance of this partial model.

        If passed object is instance of (a subclass of) this or the original model,
        no validation is performed.

        Returns partial instance with the successfully parsed fields.

        Raises ValidationError if parsing fails.
        (usless ignore_invalid is set by default or passed).
        """
        if isinstance(obj, (cls, cls.__partial_src__)):
            # safe, because subclasses are "stricter"
            return cls.construct(**obj.__dict__)  # type: ignore

        if ignore_invalid:
            # validate data and keep only valid fields
            data, fields, _ = validate_model(cls, obj)  # type: ignore
            return cls.construct(_fields_set=fields, **data)  # type: ignore

        # parse a dict or another pydantic model
        if isinstance(obj, BaseModel):
            obj = obj.dict(exclude_none=True)  # type: ignore
        return cls.parse_obj(obj)  # type: ignore

    @classmethod
    def cast(
        cls,
        obj: Union[BaseModel, PartialModel],
        *,
        ignore_invalid: bool = False,
    ):
        """Cast given object into this partial model if needed.

        If it already is an instance, will do nothing.
        Otherwise, will call `to_partial`.
        """
        if isinstance(obj, cls):
            return obj
        return cls.to_partial(obj, ignore_invalid=ignore_invalid)

    def _update_field(
        self,
        v_old,
        v_new,
        *,
        path: List[str] = None,
        allow_overwrite: bool = False,
    ):
        """Return merged result of the two passed arguments.

        None is always overwritten by a non-None value,
        lists are concatenated, sets are unionized,
        partial models are recursively merged,
        otherwise the new value overwrites the old one.
        """
        path = path or []
        # None -> missing value -> just use new value (shortcut)
        if v_old is None or v_new is None:
            return v_new or v_old

        # list -> new one must also be a list -> concatenate
        if isinstance(v_old, list):
            return v_old + v_new

        # set -> new one must also be a set -> set union
        if isinstance(v_old, set):
            # NOTE: we could try being smarter for sets of partial models
            # https://github.com/Materials-Data-Science-and-Informatics/metador-core/issues/20
            return v_old.union(v_new)  # set union

        # another model -> recursive merge of partials, if compatible
        old_is_model = isinstance(v_old, self.__partial_fac__.base_model)
        new_is_model = isinstance(v_new, self.__partial_fac__.base_model)
        if old_is_model and new_is_model:
            v_old_p = self.__partial_fac__.get_partial(type(v_old)).cast(v_old)
            v_new_p = self.__partial_fac__.get_partial(type(v_new)).cast(v_new)
            new_subclass_old = issubclass(type(v_new_p), type(v_old_p))
            old_subclass_new = issubclass(type(v_old_p), type(v_new_p))
            if new_subclass_old or old_subclass_new:
                try:
                    return v_old_p.merge_with(
                        v_new_p, allow_overwrite=allow_overwrite, _path=path
                    )
                except ValidationError:
                    # casting failed -> proceed to next merge variant
                    # TODO: maybe raise unless "ignore invalid"?
                    pass

        # if we're here, treat it as an opaque value
        if not allow_overwrite:
            msg_title = (
                f"Can't overwrite (allow_overwrite=False) at {' -> '.join(path)}:"
            )
            msg = f"{msg_title}\n\t{repr(v_old)}\n\twith\n\t{repr(v_new)}"
            raise ValueError(msg)
        return v_new

    def merge_with(
        self,
        obj,
        *,
        ignore_invalid: bool = False,
        allow_overwrite: bool = False,
        _path: List[str] = None,
    ):
        """Return a new partial model with updated fields (without validation).

        Raises `ValidationError` if passed `obj` is not suitable,
        unless `ignore_invalid` is set to `True`.

        Raises `ValueError` if `allow_overwrite=False` and a value would be overwritten.
        """
        _path = _path or []
        obj = self.cast(obj, ignore_invalid=ignore_invalid)  # raises on failure

        ret = self.copy()  # type: ignore
        for f_name, v_new in self.__partial_fac__._get_field_vals(obj):
            v_old = ret.__dict__.get(f_name)
            v_merged = self._update_field(
                v_old, v_new, path=_path + [f_name], allow_overwrite=allow_overwrite
            )
            ret.__dict__[f_name] = v_merged
        return ret

    @classmethod
    def merge(cls, *objs: PartialModel, **kwargs) -> PartialModel:
        """Merge all passed partial models in given order using `merge_with`."""
        # sadly it looks like *args and named kwargs (,*,) syntax cannot be mixed
        ignore_invalid = kwargs.get("ignore_invalid", False)
        allow_overwrite = kwargs.get("allow_overwrite", False)

        if not objs:
            return cls()

        def merge_two(x, y):
            return cls.cast(x).merge_with(
                y, ignore_invalid=ignore_invalid, allow_overwrite=allow_overwrite
            )

        return cls.cast(reduce(merge_two, objs))


# ----
# Partial factory
#
# PartialModel-specific lookup dicts
_partials: Dict[Type[PartialFactory], Dict[Type[BaseModel], Type[PartialModel]]] = {}
_forwardrefs: Dict[Type[PartialFactory], Dict[str, Type[PartialModel]]] = {}


class PartialFactory:
    """Factory class to create and manage partial models."""

    base_model: Type[BaseModel] = BaseModel
    partial_mixin: Type[PartialModel] = PartialModel

    # TODO: how to configure the whole partial family
    # with default parameters for merge() efficiently?
    # ----
    # default arguments for merge (if not explicitly passed)
    # allow_overwrite: bool = False
    # """Default argument for merge() of partials."""

    # ignore_invalid: bool = False
    # """Default argument for merge() of partials."""

    @classmethod
    def _is_base_subclass(cls, obj: Any) -> bool:
        if not isinstance(obj, type):
            return False  # not a class (probably just a hint)
        if not issubclass(obj, cls.base_model):
            return False  # not a suitable model
        return True

    @classmethod
    def _partial_name(cls, mcls: Type[BaseModel]) -> str:
        """Return class name for partial of model `mcls`."""
        return f"{mcls.__qualname__}.{cls.partial_mixin.__name__}"

    @classmethod
    def _partial_forwardref_name(cls, mcls: Type[BaseModel]) -> str:
        """Return ForwardRef string for partial of model `mcls`."""
        return f"__{mcls.__module__}_{cls._partial_name(mcls)}".replace(".", "_")

    @classmethod
    def _get_field_vals(cls, obj: BaseModel) -> Iterator[Tuple[str, Any]]:
        """Return field values, excluding None and private fields.

        This is different from `BaseModel.dict` as it ignores the defined alias
        and is used here only for "internal representation".
        """
        return (
            (k, v)
            for k, v in obj.__dict__.items()
            if is_public_name(k) and v is not None
        )

    @classmethod
    def _nested_models(cls, field_types: Dict[str, t.TypeHint]) -> Set[Type[BaseModel]]:
        """Collect all compatible nested model classes (for which we need partials)."""
        return {
            cast(Type[BaseModel], h)
            for th in field_types.values()
            for h in t.traverse_typehint(th)
            if cls._is_base_subclass(h)
        }

    @classmethod
    def _model_to_partial_fref(cls, orig_type: t.TypeHint) -> t.TypeHint:
        """Substitute type hint with forward reference to partial models.

        Will return unchanged type if passed argument is not suitable.
        """
        if not cls._is_base_subclass(orig_type):
            return orig_type
        return ForwardRef(cls._partial_forwardref_name(orig_type))

    @classmethod
    def _model_to_partial(
        cls, mcls: Type[BaseModel]
    ) -> Type[Union[BaseModel, PartialModel]]:
        """Substitute a model class with partial model.

        Will return unchanged type if argument not suitable.
        """
        return cls.get_partial(mcls) if cls._is_base_subclass(mcls) else mcls

    @classmethod
    def _partial_type(cls, orig_type: t.TypeHint) -> t.TypeHint:
        """Convert a field type hint into a type hint for the partial.

        This will make the field optional and also replace all nested models
        derived from the configured base_model with the respective partial model.
        """
        return Optional[t.map_typehint(orig_type, cls._model_to_partial_fref)]

    @classmethod
    def _partial_field(cls, orig_type: t.TypeHint) -> Tuple[Type, Optional[FieldInfo]]:
        """Return a field declaration tuple for dynamic model creation."""
        th, fi = orig_type, None

        # if pydantic Field is added (in an Annotated[...]) - unwrap
        if t.get_origin(orig_type) is Annotated:
            args = t.get_args(orig_type)
            th = args[0]
            fi = next(filter(lambda ann: isinstance(ann, FieldInfo), args[1:]), None)

        pth = cls._partial_type(th)  # map the (unwrapped) type to optional
        return (pth, fi)

    @classmethod
    def _create_base_partial(cls):
        class PartialBaseModel(cls.partial_mixin, cls.base_model):
            class Config:
                frozen = True  # make sure it's hashable

        return PartialBaseModel

    @classmethod
    def _create_partial(cls, mcls: Type[BaseModel], *, typehints=None):
        """Create a new partial model class based on `mcls`."""
        if not cls._is_base_subclass(mcls):
            raise TypeError(f"{mcls} is not subclass of {cls.base_model.__name__}!")
        if mcls is cls.base_model:
            return (cls._create_base_partial(), [])
        # ----
        # get field type annotations (or use the passed ones / for performance)
        hints = typehints or t.get_type_hints(mcls)
        field_types = {k: v for k, v in hints.items() if k in mcls.__fields__}
        # get dependencies that must be substituted
        missing_partials = cls._nested_models(field_types)
        # compute new field types
        new_fields = {k: cls._partial_field(v) for k, v in field_types.items()}
        # replace base classes with corresponding partial bases
        new_bases = tuple(map(cls._model_to_partial, mcls.__bases__))
        # create partial model
        ret: Type[PartialModel] = create_model(
            cls._partial_name(mcls),
            __base__=new_bases,
            __module__=mcls.__module__,
            __validators__=mcls.__validators__,  # type: ignore
            **new_fields,
        )
        ret.__partial_src__ = mcls  # connect to original model
        ret.__partial_fac__ = cls  # connect to this class
        # ----
        return ret, missing_partials

    @classmethod
    def get_partial(cls, mcls: Type[BaseModel], *, typehints=None):
        """Return a partial schema with all fields of the given schema optional.

        Original default values are not respected and are set to `None`.

        The use of the returned class is for validating partial data before
        zipping together partial results into a completed one.

        This is a much more fancy version of e.g.
        https://github.com/pydantic/pydantic/issues/1799

        because it recursively substitutes with partial models.
        This allows us to implement smart deep merge for partials.
        """
        if cls not in _partials:  # first use of this partial factory
            _partials[cls] = {}
            _forwardrefs[cls] = {}

        if partial := _partials[cls].get(mcls):
            return partial  # already have a partial
        else:  # block the spot (to break recursion)
            _partials[cls][mcls] = None

        # ----
        # create a partial for a model:
        # localns = {cls._partial_forwardref_name(k): v for k,v in _partials[cls].items() if v}
        localns: Dict[str, Any] = {}
        mcls.update_forward_refs(**localns)  # to be sure

        partial, nested = cls._create_partial(mcls, typehints=typehints)
        partial_ref = cls._partial_forwardref_name(mcls)
        # store result
        _forwardrefs[cls][partial_ref] = partial
        _partials[cls][mcls] = partial
        # create partials for nested models
        for model in nested:
            cls.get_partial(model)
        # resolve possible circular references
        partial.update_forward_refs(**_forwardrefs[cls])  # type: ignore
        # ----
        return partial



import pytest

from metador_core.schema.inspect import (
    UnsupportedOperation,
    WrappedLiftedDict,
    lift_dict,
)


def test_lifted_ro_dict():
    """Test LiftedRODict functionality."""
    dct = dict(x=0, y=1, z=2)
    LDct = lift_dict("LDct", dct)

    # check it's read-only (__setattr__)
    with pytest.raises(UnsupportedOperation):
        LDct.a = 5
    with pytest.raises(TypeError):
        LDct["a"] = 5

    # read (__getattr__, __getitem__)
    assert LDct.x == 0
    assert LDct["x"] == LDct.x
    with pytest.raises(AttributeError):
        LDct.a

    # __contains__
    assert "a" not in LDct
    assert "x" in LDct

    # __bool__
    assert not bool(lift_dict("LDctEmpty", {}))
    assert bool(LDct)

    # __dir__
    assert dir(LDct) == ["x", "y", "z"]

    LDct2 = lift_dict("LDct2", dct, keys=["z", "y", "x"])

    # __iter__
    assert list(iter(LDct)) == ["x", "y", "z"]
    assert list(iter(LDct2)) == ["z", "y", "x"]

    # __repr__
    r1 = repr(LDct)
    assert 0 <= r1.find("x") < r1.find("y") < r1.find("z")
    r2 = repr(LDct2)
    assert 0 <= r2.find("z") < r2.find("y") < r2.find("x")


def test_wrapped():
    """Test WrappedLiftedDict (with a value mapping function)."""
    LDct = lift_dict("LDct", dict(hello="world"))
    WDct = WrappedLiftedDict(LDct, lambda _: "surprise")
    assert WDct.hello == "surprise"
    assert WDct["hello"] == "surprise"
    assert repr(LDct) == repr(WDct)



"""Test skeleton and stub creation (decoupled from manifest file)."""
import pytest

from metador_core.ih5.container import IH5Record
from metador_core.ih5.skeleton import (
    H5Type,
    IH5Skeleton,
    SkeletonNodeInfo,
    init_stub_base,
    init_stub_skeleton,
)


def test_ih5_skeleton(tmp_ds_path_factory):
    with IH5Record(tmp_ds_path_factory(), "w") as ds:
        ds["foo/bar"] = "hello"
        ds.attrs["root_attr"] = 1
        ds["foo"].attrs["group_attr"] = 2
        ds.commit_patch()
        ds.create_patch()
        ds["foo"].attrs["group_attr2"] = 3
        ds["foo/bar"].attrs["dataset_attr"] = 4  # type: ignore
        ds["foo/baz"] = "world"
        ds["foo/baz"].attrs["dataset_attr2"] = 5  # type: ignore

        assert IH5Skeleton.for_record(ds) == IH5Skeleton(
            __root__={
                "/": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=0, attrs={"root_attr": 0}
                ),
                "/foo": SkeletonNodeInfo(
                    node_type=H5Type.group,
                    patch_index=0,
                    attrs={"group_attr": 0, "group_attr2": 1},
                ),
                "/foo/bar": SkeletonNodeInfo(
                    node_type=H5Type.dataset, patch_index=0, attrs={"dataset_attr": 1}
                ),
                "/foo/baz": SkeletonNodeInfo(
                    node_type=H5Type.dataset, patch_index=1, attrs={"dataset_attr2": 1}
                ),
            }
        )


def test_init_stub_skeleton(tmp_ds_path_factory):
    with IH5Record(tmp_ds_path_factory(), "w") as ds:
        skel = {  # dict in key creation order. Key order should not matter!
            "foo": {"node_type": "dataset", "patch_index": 0, "attrs": {"atr": 0}},
            "qux": {
                "node_type": "group",
                "patch_index": 0,
                "attrs": {},
            },
            "qux/bar": {
                "node_type": "dataset",
                "patch_index": 0,
                "attrs": {},
            },
        }
        # in this order it still should work fine
        init_stub_skeleton(ds, IH5Skeleton.parse_obj(skel))

    # now test failures
    with IH5Record(tmp_ds_path_factory(), "w") as ds:
        # an invalid skeleton
        with pytest.raises(ValueError):
            init_stub_skeleton(ds, IH5Skeleton.parse_obj({"foo": "invalid"}))

    with IH5Record(tmp_ds_path_factory(), "w") as ds:
        # not empty, but no collision
        ds["bar"] = "not empty anymore"
        with pytest.raises(ValueError):
            init_stub_skeleton(ds, IH5Skeleton.parse_obj({"foo": "dataset"}))


def test_patch_on_stub_works_with_real(tmp_ds_path_factory):
    # Test that patching over a stub works correctly and can be used with original.

    # create a little normal record with multiple patches
    dsname = tmp_ds_path_factory()
    stubname = tmp_ds_path_factory()
    with IH5Record(dsname, "w") as ds:
        ds["foo/bar"] = [1, 2, 3]
        ds.commit_patch()
        ds.create_patch()
        ds["foo/bar"].attrs["qux"] = 42  # type: ignore
        ds["data"] = "interesting data"
        ds.commit_patch()
        ds.create_patch()
        ds["data"].attrs["key"] = "value"  # type: ignore
        ds["foo/muh"] = 1337
        ds["foo/tokill"] = "this will be deleted"
        ds.commit_patch()

        ds_files = ds.ih5_files
        ds_ub = ds.ih5_meta[-1]
        ds_sk = IH5Skeleton.for_record(ds)

        # create the stub faking this record
        with IH5Record(stubname, "w") as stub:
            init_stub_base(stub, ds_ub, ds_sk)
            stub.commit_patch()

            # should agree on relevant user block infos
            assert ds_ub.record_uuid == stub.ih5_meta[0].record_uuid
            assert ds_ub.patch_uuid == stub.ih5_meta[0].patch_uuid
            assert ds_ub.patch_index == stub.ih5_meta[0].patch_index
            # and of course on the skeleton
            latest_pidx = stub.ih5_meta[-1].patch_index
            assert IH5Skeleton.for_record(stub) == ds_sk.with_patch_index(latest_pidx)

            # create new patch on top of stub
            stub.create_patch()
            del stub["foo/tokill"]
            del stub["foo/bar"]
            stub["/foo/bar/blub"] = True
            stub["data"].attrs["key2"] = "othervalue"
            stub["foo/bar"].attrs["qax"] = 987  # type: ignore
            stub.commit_patch()

            assert len(stub.ih5_files) == 2
            stub_files = stub.ih5_files
            stub_skel = IH5Skeleton.for_record(stub)
            assert stub_skel == IH5Skeleton.parse_obj(
                {
                    "/": {"node_type": "group", "patch_index": 2, "attrs": {}},
                    "/data": {
                        "node_type": "dataset",
                        "patch_index": 2,
                        "attrs": {"key": 2, "key2": 3},
                    },
                    "/foo": {"node_type": "group", "patch_index": 2, "attrs": {}},
                    "/foo/bar": {
                        "node_type": "group",
                        "patch_index": 3,
                        "attrs": {"qax": 3},
                    },
                    "/foo/bar/blub": {
                        "node_type": "dataset",
                        "patch_index": 3,
                        "attrs": {},
                    },
                    "/foo/muh": {"node_type": "dataset", "patch_index": 2, "attrs": {}},
                }
            )

    # open real record with new patch in stub
    with IH5Record._open([*ds_files, stub_files[-1]]) as ds:
        # first success is that it even opens without complaining.
        # now check that it's the same skeleton with the real data:
        # (ignoring patch indices, which are different)
        ds_skel = IH5Skeleton.for_record(ds)
        assert ds_skel.with_patch_index(0) == stub_skel.with_patch_index(0)
        # check the attributes are merged and values look as expeted
        assert set(ds["data"].attrs.keys()) == set(["key", "key2"])
        assert set(ds["foo/bar"].attrs.keys()) == set(["qax"])
        assert "foo/muh" in ds
        assert "foo/bar/blub" in ds



"""Test hashing helper functions."""
import pytest

from metador_core.util.hashsums import file_hashsum


def test_hashsum(tmp_path):
    file = tmp_path / "test.txt"
    with open(file, "w") as f:
        f.write("hello world!")

    with pytest.raises(ValueError):
        file_hashsum(file, "invalid")

    hsum = file_hashsum(file, "sha256")
    assert (
        hsum
        == "sha256:7509e5bda0c762d2bac7f90d758b5b2263fa01ccbc542ab5e3df163be08e6ca9"
    )



"""Metador interface to manage metadata in HDF5 containers.

Works with plain h5py.File and IH5Record subclasses and can be extended to work
with any type of archive providing the required functions in a h5py-like interface.

When assembling a container, the compliance with the Metador container specification is
ensured by using it through the MetadorContainer interface.

Technical Metador container specification (not required for users):

Metador uses only HDF5 Groups and Datasets. We call both kinds of objects Nodes.
Notice that HardLinks, SymLinks, ExternalLinks or region references cannot be used.

Users are free to lay out data in the container as they please, with one exception:
a user-defined Node MUST NOT have a name starting with "metador_".
"metador_" is a reserved prefix for Group and Dataset names used to manage
technical bookkeeping structures that are needed for providing all container features.

For each HDF5 Group or Dataset there MAY exist a corresponding
Group for Metador-compatible metadata that is prefixed with "metador_meta_".

For "/foo/bar" the metadata is to be found...
    ...in a group "/foo/metador_meta_bar", if "/foo/bar" is a dataset,
    ...in a group "/foo/bar/metador_meta_" if it is a group.
We write meta("/foo/bar") to denote that group.

Given schemas with entrypoint names X, Y and Z such that X is the parent schema of Y,
and Y is the parent schema of Z and a node "/foo/bar" annotated by a JSON object of
type Z, that JSON object MUST be stored as a newline-terminated, utf-8 encoded byte
sequence at the path meta("/foo/bar")/X/Y/Z/=UUID, where the UUID is unique in the
container.

For metadata attached to an object we expect the following to hold:

Node instance uniqueness:
Each schema MAY be instantiated explicitly for each node at most ONCE.
Collections thus must be represented on schema-level whenever needed.

Parent Validity:
Any object of a subschema MUST also be a valid instance of all its parent schemas.
The schema developers are responsible to ensure this by correct implementation
of subschemas.

Parent Consistency:
Any objects of a subtype of schema X that stored at the same node SHOULD result
in the same object when parsed as X (they agree on the "common" information).
Thus, any child object can be used to retrieve the same parent view on the data.
The container creator is responsible for ensuring this property. In case it is not
fulfilled, retrieving data for a more abstract type will yield it from ANY present subtype
instance (but always the same one, as long as the container does not change)!

If at least one metadata object it stored, a container MUST have a "/metador_toc" Group,
containing a lookup index of all metadata objects following a registered metadata schema.
This index structure MUST be in sync with the node metadata annotations.
Keeping this structure in sync is responsibility of the container interface.

This means (using the previous example) that for "/foo/bar" annotated by Z there also
exists a dataset "/metador_toc/X/Y/Z/=UUID" containing the full path to the metadata node,
i.e. "meta(/foo/bar)/X/Y/Z/=UUID". Conversely, there must not be any empty entry-point
named Groups, and all listed paths in the TOC must point to an existing node.

A valid container MUST contain a dataset /metador_version string of the form "X.Y"

A correctly implemented library supporting an older minor version MUST be able open a
container with increased minor version without problems (by ignoring unknown data),
so for a minor update of this specification only new entities may be defined.

Known technical limitations:

Due to the fact that versioning of plugins such as schemas is coupled to the versioning
of the respective Python packages, it is not (directly) possible to use two different
versions of the same schema in the same environment (with the exception of mappings, as
they may bring their own equivalent schema classes).

Minor version updates of packages providing schemas must ensure that the classes providing
schemas are backward-compatible (i.e. can parse instances of older minor versions).

Major version updates must also provide mappings migrating between the old and new schema
versions. In case that the schema did not change, the mapping is simply the identity.
"""

from .interface import MetadorContainerTOC, MetadorMeta
from .provider import ContainerProxy
from .wrappers import MetadorContainer, MetadorDataset, MetadorGroup, MetadorNode

__all__ = [
    "MetadorContainer",
    "MetadorNode",
    "MetadorGroup",
    "MetadorDataset",
    "MetadorMeta",
    "MetadorContainerTOC",
    "ContainerProxy",
]



from __future__ import annotations

from typing import Any, Dict, List, Optional


class DirValidationErrors(ValueError):
    """Structure to collect record or directory validation errors.

    Returned from validation functions and thrown in other contexts.
    """

    errors: Dict[str, List[Any]]
    """Common type used to collect errors.

    Maps path in container or directory to list of errors with that path.

    The list should contain either strings or other such dicts,
    but Python type checkers are unable to understand recursive types.
    """

    def __init__(self, errs: Optional[Dict[str, Any]] = None):
        self.errors = errs or {}

    def __bool__(self):
        return bool(self.errors)

    def __repr__(self):  # pragma: no cover
        return repr(self.errors)

    def update(self, *err_objs: DirValidationErrors):
        """Add errors from other instances to this object."""
        errs = self.errors
        for more_errs in err_objs:
            for k, v in more_errs.errors.items():
                if k not in errs:
                    errs[k] = v
                else:
                    errs[k] += v

    def add(self, k, v):
        if k not in self.errors:
            self.errors[k] = []
        self.errors[k].append(v)



from hypothesis import given
from hypothesis import strategies as st

from metador_core.container.utils import (
    is_internal_path,
    is_meta_base_path,
    to_data_node_path,
    to_meta_base_path,
)


def test_is_internal_path():
    assert not is_internal_path("test")
    assert not is_internal_path("_metador_something")
    assert not is_internal_path("/_metador_something")
    assert not is_internal_path("/test/_metador_something")
    assert is_internal_path("metador_something")
    assert is_internal_path("/metador_something")
    assert is_internal_path("/test/metador_something_else/")

    assert is_internal_path("/a/prefix_something/b", "prefix_")
    assert not is_internal_path("/a/_prefix_something/b", "prefix_")


def test_is_meta_base_path():
    assert not is_meta_base_path("_metador_meta_b")
    assert is_meta_base_path("metador_meta_")
    assert is_meta_base_path("/metador_meta_a")
    assert is_meta_base_path("/a/metador_meta_b")


def test_to_meta_base_path():
    assert to_meta_base_path("/", False) == "/metador_meta_"
    assert to_meta_base_path("/", True) == "/metador_meta_"
    assert to_meta_base_path("/a/b", False) == "/a/b/metador_meta_"
    assert to_meta_base_path("/a/b", True) == "/a/metador_meta_b"


def test_to_data_node_path():
    assert to_data_node_path("/metador_meta_") == "/"
    assert to_data_node_path("/metador_meta_a") == "/a"
    assert to_data_node_path("a/metador_meta_b") == "a/b"
    assert to_data_node_path("a/b/metador_meta_") == "a/b"


# NOTE: 0-~ excludes /, but includes most normal ASCII printable characters
rel_paths = st.from_regex(r"([0-~]+)(/[0-~]+)*", fullmatch=True)
abs_paths = st.one_of(st.sampled_from("/"), rel_paths.map(lambda s: f"/{s}"))
paths = st.one_of(abs_paths, rel_paths)


@given(st.tuples(paths, st.booleans()))
def test_meta_base_data_node_inverses(node):
    """Test that converting a path into the metadata directory path and back is correct."""
    n_path, n_type = node
    meta_path = to_meta_base_path(n_path, n_type)
    data_path = to_data_node_path(meta_path)
    assert n_path == data_path



from __future__ import annotations

import itertools
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Union

from pydantic import BaseModel

from .hashsums import DirHashsums


def dir_paths(base_dir: Path):
    """Recursively list all paths in given directory, relative to itself."""
    return map(lambda p: p.relative_to(base_dir), sorted(base_dir.rglob("*")))


class DiffNode(BaseModel):
    """Node representing a file, symlink or directory in a diff."""

    class Status(str, Enum):
        """Change that happened to a DiffNode."""

        removed = "-"
        modified = "~"
        added = "+"
        unchanged = "0"

    class ObjType(str, Enum):
        """Entities represented in a DiffNode."""

        directory = "d"
        file = "f"
        symlink = "s"

    path: Path
    """Location represented by this node."""

    prev: Union[None, str, Dict[str, Any]]
    """Previous entity at this location."""

    curr: Union[None, str, Dict[str, Any]]
    """Current entity at this location."""

    removed: Dict[Path, DiffNode] = {}
    """Deleted files and subdirectories."""

    modified: Dict[Path, DiffNode] = {}
    """Modified or replaced files and subdirectories."""

    added: Dict[Path, DiffNode] = {}
    """New files and subdirectories."""

    def _type(self, entity) -> Optional[DiffNode.ObjType]:
        if isinstance(entity, dict):
            return DiffNode.ObjType.directory
        elif entity:
            if entity.find("symlink:") == 0:
                return DiffNode.ObjType.symlink
            else:
                return DiffNode.ObjType.file
        return None

    @property
    def prev_type(self) -> Optional[DiffNode.ObjType]:
        return self._type(self.prev)

    @property
    def curr_type(self) -> Optional[DiffNode.ObjType]:
        return self._type(self.curr)

    def children(self) -> Iterable[DiffNode]:
        """Return immediate children in no specific order."""
        return itertools.chain(
            *(x.values() for x in [self.removed, self.modified, self.added])
        )

    def nodes(self) -> List[DiffNode]:
        """Return list of all nodes in this diff.

        The order will be recursively:
        removed children, modified children, itself, then added children.
        Within the same category the children are sorted in alphabetical order.
        """
        ret = []
        buckets = [self.removed, self.modified, None, self.added]
        for b in buckets:
            if b is None:
                ret.append(self)
            else:
                for v in sorted(b.values(), key=lambda x: x.path):
                    ret += v.nodes()
        return ret

    def status(self) -> DiffNode.Status:
        """Check the given path (which is assumed to be relative to this diff node).

        Returns whether the path was added, removed or modified, respectively.
        """
        if self.prev is None:
            return DiffNode.Status.added
        elif self.curr is None:
            return DiffNode.Status.removed
        else:  # we can assume self.prev != self.curr
            return DiffNode.Status.modified

    @classmethod
    def compare(
        cls,
        prev: Optional[DirHashsums],
        curr: Optional[DirHashsums],
        path: Path,
    ) -> Optional[DiffNode]:
        """Compare two nested file and directory hashsum dicts.

        Returns None if no difference is found, otherwise a DiffNode tree
        containing only the additions, removals and changes.
        """
        ret = cls(path=path, prev=prev, curr=curr)
        if not isinstance(prev, dict) and not isinstance(curr, dict):
            if prev == curr:
                return None  # same (non-)file -> no diff
            else:
                return ret  # indicates that a change happened

        if (prev is None or isinstance(prev, str)) and isinstance(curr, dict):
            # file -> dir: everything inside "added"
            for k, v in curr.items():
                kpath = ret.path / k
                d = cls.compare(None, v, kpath)
                assert d is not None
                ret.added[kpath] = d
            return ret

        if isinstance(prev, dict) and (curr is None or isinstance(curr, str)):
            # dir -> file: everything inside "removed"
            for k, v in prev.items():
                kpath = ret.path / k
                d = cls.compare(v, None, kpath)
                assert d is not None
                ret.removed[kpath] = d
            return ret

        assert isinstance(prev, dict) and isinstance(curr, dict)
        # two directories -> compare
        prev_keys = set(prev.keys())
        curr_keys = set(curr.keys())

        added = curr_keys - prev_keys
        removed = prev_keys - curr_keys
        intersection = (prev_keys | curr_keys) - added - removed

        for k in added:  # added in curr
            kpath = ret.path / k
            d = cls.compare(None, curr[k], kpath)
            assert d is not None
            ret.added[kpath] = d
        for k in removed:  # removed in curr
            kpath = ret.path / k
            d = cls.compare(prev[k], None, kpath)
            assert d is not None
            ret.removed[kpath] = d
        for k in intersection:  # changed in curr
            kpath = ret.path / k
            diff = cls.compare(prev[k], curr[k], kpath)
            if diff is not None:  # add child if there is a difference
                ret.modified[kpath] = diff

        # all children same -> directories same
        same_dir = not ret.added and not ret.removed and not ret.modified
        if same_dir:
            return None

        # return dir node with the changes
        return ret


class DirDiff:
    """Interface to directory diffs based on comparing `DirHashsums`.

    An instance represents a change at the path.

    Granular change inspection is accessible through the provided methods.

    Typically, you will want to capture `dir_hashsums` of the same `dir`
    at two points in time and will be interested in the dict
    `DirDiff.compare(prev, curr).annotate(dir)`.
    """

    _diff_root: Optional[DiffNode]

    @property
    def is_empty(self):
        return self._diff_root is None

    @classmethod
    def compare(cls, prev: DirHashsums, curr: DirHashsums) -> DirDiff:
        """Compute a DirDiff based on two DirHashsum trees.

        To be meaningful, `prev` and `curr` should be trees obtained from the
        same directory at two points in time.
        """
        ret = cls.__new__(cls)
        ret._diff_root = DiffNode.compare(prev, curr, Path(""))
        return ret

    def get(self, path: Path) -> Optional[DiffNode]:
        path = Path(path)  # if it was a str
        assert not path.is_absolute()
        if self._diff_root is None:
            return None
        curr: DiffNode = self._diff_root
        prefixes = [path] + list(path.parents)
        prefixes.pop()  # drop '.'
        while prefixes:
            path = prefixes.pop()
            node: Optional[DiffNode] = next(
                (x for x in curr.children() if x.path == path), None
            )
            if node is None:
                return None
            curr = node
        return curr

    def status(self, node: Optional[DiffNode]) -> DiffNode.Status:
        """Return the type of change that happened to a diff node.

        Wraps `node.status()` additionally covering the case that `node` is `None`.
        Useful when processing possible path nodes returned by `annotate`.
        """
        if node is None:
            return DiffNode.Status.unchanged
        return node.status()

    def annotate(self, base_dir: Path) -> Dict[Path, Optional[DiffNode]]:
        """Return a dict of path -> status mappings based on passed directory.

        The keys are all paths that exist in this diff object as well as
        all paths that currently exist in the passed `base_dir`.
        The keys will all have `base_dir` as prefix.

        The values will be the corresponding DiffNodes, if there is a change.
        If there was no change at that path, the value will be None.

        The iteration order will be such that for each entity, first the removed
        children, then changed subentities, then the entity itself, and finally
        added subentities are listed. This is useful for updating operations
        of directories and ensures that children can be deleted before their parents
        and parents can be created before their children.
        """
        if self._diff_root is None:
            return {}

        nodes = self._diff_root.nodes()  # nodes of all paths in the diff

        path_nodes = {node.path: node for node in nodes}
        # paths in base_dir, but not in the diff
        missing = sorted(set(dir_paths(base_dir)) - set(path_nodes.keys()))

        # construct result, keeping correct order and prepending base_dir
        ret: Dict[Path, Optional[DiffNode]] = {
            base_dir / str(k): v for k, v in path_nodes.items()
        }
        for path in missing:
            ret[base_dir / str(path)] = None
        return ret



from __future__ import annotations

import itertools
from enum import Enum, EnumMeta
from typing import Any, Dict, Optional, Set, Union


class NestedEnumMeta(EnumMeta):
    """Enum subclass for Enums generated from a semantic taxonomy.

    Approach is to collect values and treat them as sets.
    """

    __self__: Optional[Union[int, str]] = None
    __values__: Set[Any] = set()
    __nested__: Dict[str, NestedEnumMeta] = {}

    name: str
    value: Union[int, str]

    def __call__(metacls, clsname, dct, *args, **kwargs):
        self = kwargs.pop("self", None)
        mod_dct = {}
        nested = {}
        for k, v in dct.items():
            if isinstance(v, NestedEnumMeta):
                if v.__self__ is None:
                    raise ValueError(f"Nested enum {v} must have a self value!")
                nested[k] = v
            else:
                mod_dct[k] = v

        ret = super().__call__(clsname, mod_dct, *args, **kwargs)

        ret.__self__ = self
        nested_values = set.union(
            set(), *map(lambda x: {x.__self__}.union(x.__values__), nested.values())
        )
        ret.__values__ = set(map(lambda x: x.value, iter(ret))).union(nested_values)
        ret.__nested__ = nested
        for name, nst in nested.items():
            setattr(ret, name, nst)

        return ret

    def __contains__(self, other):
        print(other, "in", self, "?")
        if isinstance(other, NestedEnum):
            return self.__contains__(other.value)
        if isinstance(other, type) and issubclass(other.__class__, NestedEnumMeta):
            return self.__contains__(other.__self__)
        # lookup plain value
        return other in self.__values__

    def __iter__(self):
        return itertools.chain(
            super().__iter__(),
            *(itertools.chain(iter((x,)), iter(x)) for x in self.__nested__.values()),
        )

    def __dir__(self):
        return itertools.chain(super().__dir__(), self.__nested__.keys())

    def __repr__(self):
        if self.__self__ is not None:
            return f"<enum {self.__name__}: {self.__self__}>"
        else:
            return super().__repr__()


class NestedEnum(Enum, metaclass=NestedEnumMeta):
    """Base class for hierarchical enumerations."""



"""General utitilies with relevance for plugins."""
from typing import TYPE_CHECKING, Any, Optional, Type, TypeVar

from ..util import eprint
from .types import PluginLike, is_pluginlike, to_ep_name

if TYPE_CHECKING:  # pragma: no cover
    from .interface import PluginGroup
else:
    PluginGroup = Any


# ----
# helpers for checking plugins (also to be used in PluginGroup subclasses):


def implements_method(plugin, base_method):
    ep_method = plugin.__dict__.get(base_method.__name__)
    return ep_method is not None and base_method != ep_method


def check_implements_method(name: str, plugin, base_method):
    """Check whether plugin overrides a method of its superclass."""
    if not implements_method(plugin, base_method):
        msg = f"{name}: {plugin} does not implement {base_method.__name__}!"
        raise TypeError(msg)


def check_is_subclass(name: str, plugin, base):
    """Check whether plugin has expected parent class (helper method)."""
    if not issubclass(plugin, base):
        msg = f"{name}: {plugin} is not subclass of {base}!"
        raise TypeError(msg)


# ----


def is_notebook() -> bool:  # pragma: no cover
    # https://stackoverflow.com/a/39662359
    try:
        # get_ipython() is defined globally in ipython-like env!
        shell = get_ipython().__class__.__name__  # type: ignore

        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


T = TypeVar("T", bound=PluginLike)


def register_in_group(
    pgroup: PluginGroup,
    plugin: Optional[Type[T]] = None,
    *,
    violently: bool = False,
):
    """Register and load a plugin manually, without defining an entry point."""
    if not violently and not is_notebook():
        raise RuntimeError("This is not supposed to be used outside of notebooks!")

    def manual_register(plugin: Type[T]) -> Type[T]:
        pginfo = plugin.Plugin
        ep_name = to_ep_name(pginfo.name, pginfo.version)
        pg_ref = pgroup.PluginRef(name=pginfo.name, version=pginfo.version)

        pgroup._ENTRY_POINTS[ep_name] = None
        pgroup._LOADED_PLUGINS[pg_ref] = plugin
        if pg_ref.name not in pgroup._VERSIONS:
            pgroup._VERSIONS[pg_ref.name] = []
        pgroup._VERSIONS[pg_ref.name].append(pg_ref)

        pgroup._load_plugin(ep_name, plugin)
        if not violently:
            eprint(
                f"Notebook: Plugin '{pginfo.name}' registered in '{pgroup.name}' group!"
            )  # pragma: no cover
        return plugin

    if not plugin:
        return manual_register  # used as decorator
    else:
        if not is_pluginlike(plugin, check_group=False):
            raise RuntimeError("This class has no inner Plugin class!")

        manual_register(plugin)  # used as normal function



import pytest

from metador_core.container import MetadorContainer
from metador_core.container.provider import SimpleContainerProvider


def test_simple_container_provider(tmp_ds_path):
    scp = SimpleContainerProvider()

    assert scp.get("test") is None
    assert "test" not in scp
    with pytest.raises(KeyError):
        scp["test"]

    tmp_ds_path.mkdir()
    filepath = tmp_ds_path / "container.h5"
    with MetadorContainer(filepath, "w") as m:
        m["hello"] = b"world"

        # add to provider
        scp["test"] = m

    # access
    assert "test" in scp
    assert "test" in scp.keys()
    with scp["test"] as m:
        assert m["hello"][()] == b"world"

    # remove
    del scp["test"]
    assert "test" not in scp



"""Common metadata models based on RO-Crate and Schema.org."""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple

from pydantic import parse_obj_as

from ...plugins import schemas
from .. import MetadataSchema
from ..decorators import make_mandatory
from ..parser import BaseParser
from ..types import PintQuantity, PintUnit
from .rocrate import Person
from .schemaorg import Number, QuantitativeValue, Text

# ----


class SIValueParser(BaseParser):
    @classmethod
    def parse(cls, tcls, v) -> SIValue:
        if isinstance(v, str):
            q = parse_obj_as(PintQuantity, v)
            return tcls(value=q.m, unitText=str(q.u))

        # special case: its a quantitative value already,
        # just check unit and repack as SIValue
        if isinstance(v, tcls.__base__):
            v.unitText = str(parse_obj_as(PintUnit, v.unitText or ""))
            return tcls.construct(v.dict())

        # important: parse back serialized data!
        if isinstance(v, dict):
            return tcls.validate(v)

        msg = f"Cannot parse {v} ({type(v).__name__}) into a {tcls.__name__}!"
        raise TypeError(msg)


class SIValue(QuantitativeValue):
    """QuantitativeValue that holds a numerical value given in SI units.

    Uses the `pint` library to parse and normalize the unit.
    """

    value: Number
    Parser = SIValueParser


class NumValue(QuantitativeValue):
    """Quantitative value that can have a unit out of a fixed list."""

    value: Number

    class Parser(BaseParser):
        schema_info: Dict[str, Any] = {}

        allowed_units: List[str] = []
        infer_unit: Optional[str] = None
        require_unit: bool = False

        @classmethod
        def parse(cls, tcls, v):
            if isinstance(v, (int, float)):
                if cls.require_unit:
                    raise ValueError(f"Value '{v}' must have a unit!")
                return tcls.construct(value=v, unitText=cls.infer_unit)

            if isinstance(v, str):
                arr = v.strip().split(maxsplit=1)

            # important to parse back serialized data!
            if isinstance(v, dict):
                v = tcls.__base__.validate(v)  # -> QuantitativeValue

            if isinstance(v, tcls.__base__):  # unpack QuantitativeValue
                def_unit = cls.infer_unit or ""
                arr = (v.value, v.unitText or v.unitCode or def_unit)

            # check that value and unit are valid:

            if len(arr) == 1:  # no unit given?
                if cls.require_unit:
                    raise ValueError(f"Value '{v}' must have a unit!")
                val = parse_obj_as(Number, arr[0])
                # return with inferred unit
                return tcls.construct(value=val, unitText=cls.infer_unit)

            val = parse_obj_as(Tuple[Number, str], arr)
            if cls.allowed_units and val[1] not in cls.allowed_units:
                msg = (
                    f"Invalid unit '{val[1]}', unit must be one of {cls.allowed_units}"
                )
                raise ValueError(msg)
            return tcls.construct(value=val[0], unitText=val[1])


class Pixels(NumValue):
    """Numeric value representing pixels."""

    class Parser(NumValue.Parser):
        allowed_units = ["px"]
        infer_unit = "px"


# ----

FileMeta: Any = schemas.get("core.file", (0, 1, 0))
DirMeta: Any = schemas.get("core.dir", (0, 1, 0))


@make_mandatory("name", "abstract", "dateCreated")
class BibMeta(DirMeta):
    """Minimal bibliographic metadata required for a container."""

    class Plugin:
        name = "core.bib"
        version = (0, 1, 0)

    author: List[Person]
    """List of authors (creators of the actual content)."""

    creator: Person
    """Person who created the container."""


class ImageFileMeta(FileMeta):
    """A rasterized image file with known dimensions.

    Also serves as marker schema for the imagefile widget.
    """

    class Plugin:
        name = "core.imagefile"
        version = (0, 1, 0)

    width: Pixels
    """Width of the image in pixels."""

    height: Pixels
    """Height of the image in pixels."""


# TODO: see https://github.com/Materials-Data-Science-and-Informatics/metador-core/issues/8


class ColumnHeader(MetadataSchema):
    """Table column metadata."""

    name: Text
    """Column title."""

    unit: PintUnit
    """Physical unit for this column."""


class TableMeta(MetadataSchema):
    """Table metadata."""

    class Plugin:
        name = "core.table"
        version = (0, 1, 0)

    name: Text
    """Table title."""

    columns: List[ColumnHeader]
    """List of column descriptions."""



"""Simple types and utilities for plugins."""
from typing import Any, ClassVar, Optional, Protocol, Tuple, runtime_checkable

from phantom.re import FullMatch
from pydantic import NonNegativeInt
from typing_extensions import Final, TypeAlias

# ----

SemVerTuple: TypeAlias = Tuple[NonNegativeInt, NonNegativeInt, NonNegativeInt]
"""Type to be used for SemVer triples."""

DIGIT: Final[str] = r"[0-9]"  # python 3: \d matches also weird other symbols!!!
SEMVER_STR_REGEX: Final[str] = rf"{DIGIT}+\.{DIGIT}+\.{DIGIT}+"


class SemVerStr(FullMatch, pattern=SEMVER_STR_REGEX):
    """Semantic version string (x.y.z)."""


def to_semver_str(ver: SemVerTuple):
    return ".".join(map(str, ver))


def from_semver_str(ver: SemVerStr) -> SemVerTuple:
    return tuple(map(int, ver.split(".")))  # type: ignore


# ----

LETTER: Final[str] = r"[a-z]"
ALNUM: Final[str] = r"[a-z0-9]"
LETSEP: Final[str] = r"[_-]"
NSSEP: Final[str] = r"[.]"
NAME: Final[str] = rf"{LETTER}{ALNUM}({LETSEP}?{ALNUM})*"
"""An unqualified name begins with a letter and ends with a letter or number.

It consists of: lowercase letters, digits, _ and -
"""

QUAL_NAME: Final[str] = rf"{NAME}({NSSEP}{NAME})*"
"""A qualified name is a sequence of unqualified names separated with ."""

EP_NAME_VER_SEP: str = "__"
"""Separator between plugin name and semantic version in entry point name."""

EP_NAME_REGEX: Final[str] = rf"{QUAL_NAME}{EP_NAME_VER_SEP}{SEMVER_STR_REGEX}"
"""Regular expression that all metador plugin entry points must match."""


class EPName(FullMatch, pattern=EP_NAME_REGEX):
    """Valid entry point for a metador plugin."""


def to_ep_name(p_name: str, p_version: SemVerTuple) -> EPName:
    """Return canonical entrypoint name `PLUGIN_NAME__MAJ.MIN.FIX`."""
    return EPName(f"{p_name}{EP_NAME_VER_SEP}{to_semver_str(p_version)}")


def from_ep_name(ep_name: EPName) -> Tuple[str, SemVerTuple]:
    """Split entrypoint name into `(PLUGIN_NAME, (MAJ,MIN,FIX))`."""
    pname, pverstr = ep_name.split(EP_NAME_VER_SEP)
    return (pname, from_semver_str(SemVerStr(pverstr)))


def ep_name_has_namespace(ep_name: EPName):
    """Check whether the passed name has a namespace prefix."""
    name, ver = from_ep_name(ep_name)
    return len(name.split(".", 1)) > 1


PG_PREFIX: str = "metador_"
"""Group prefix for metador plugin entry point groups."""


class EPGroupName(FullMatch, pattern=rf"{PG_PREFIX}.*"):
    """Valid internal group name for a metador plugin group."""


def is_metador_ep_group(ep_group_name: str):
    return ep_group_name.startswith(PG_PREFIX)


def to_ep_group_name(group_name: str) -> EPGroupName:
    return EPGroupName(f"{PG_PREFIX}{group_name}")


def from_ep_group_name(ep_group_name: EPGroupName) -> str:
    plen = len(PG_PREFIX)
    return ep_group_name[plen:]


# ----


class PluginLike(Protocol):
    """A Plugin has a Plugin inner class with plugin infos."""

    Plugin: ClassVar[Any]  # actually its PluginBase, but this happens at runtime


@runtime_checkable
class HasNameVersion(Protocol):
    name: Any
    version: Any


@runtime_checkable
class PluginInfoLike(HasNameVersion, Protocol):
    group: Any


def is_pluginlike(cls, *, check_group=True) -> bool:
    if pgi := cls.__dict__.get("Plugin"):
        return isinstance(pgi, PluginInfoLike if check_group else HasNameVersion)
    else:
        return False


def plugin_args(
    plugin="",  # actually takes: Union[str, PluginRef, PluginLike]
    version: Optional[SemVerTuple] = None,
    *,
    require_version: bool = False,
    # group: Optional[str]
) -> Tuple[str, Optional[SemVerTuple]]:
    """Return requested plugin name and version based on passed arguments.

    Helper for function argument parsing.
    """
    name: str = ""
    vers: Optional[SemVerTuple] = version

    if isinstance(plugin, str):
        name = plugin
    if isinstance(plugin, tuple) and len(plugin) == 2:
        name = plugin[0]
        if not vers:
            vers = plugin[1]
    elif isinstance(plugin, HasNameVersion):
        name = plugin.name
        if not vers:
            vers = plugin.version
    elif pgi := getattr(plugin, "Plugin", None):
        if isinstance(pgi, HasNameVersion):
            return plugin_args(pgi, version, require_version=require_version)

    if require_version and vers is None:
        raise ValueError(f"No version of {name} specified, but is required!")
    return (name, vers)



import re
import sys
from functools import lru_cache
from typing import Iterable

cache = lru_cache(maxsize=None)  # for 3.8 compat


def eprint(*args, **kwargs):
    """Print to error stream."""
    print(*args, file=sys.stderr, **kwargs)


def drop(n: int, it: Iterable):
    """Drop fixed number of elements from iterator."""
    return (x for i, x in enumerate(it) if i >= n)


def is_public_name(n: str):
    """Return whether a name is public (does not start with _)."""
    return n[0] != "_"


def pythonize_name(name: str):
    """Sanitize a string to be a valid Python variable name."""
    return re.sub(r"\W|^(?=\d)", "_", " ".join(name.split()))



"""Schemas needed for the plugin system."""
from __future__ import annotations

import json
from collections import ChainMap
from functools import total_ordering
from typing import ClassVar, Dict, List, Literal, Optional

from pydantic import AnyHttpUrl, Extra, ValidationError, create_model

from ..util import is_public_name
from .core import BaseModelPlus, MetadataSchema
from .types import NonEmptyStr, SemVerTuple, to_semver_str


@total_ordering
class PluginRef(MetadataSchema):
    """Reference to a metador plugin.

    This class can be subclassed for specific "marker schemas".

    It is not registered as a schema plugin, because it is too general on its own.
    """

    class Config:
        extra = Extra.forbid
        allow_mutation = False

    group: NonEmptyStr
    """Metador pluggable group name, i.e. name of the entry point group."""

    name: NonEmptyStr
    """Registered entry point name inside an entry point group."""

    version: SemVerTuple
    """Version of the Python package."""

    def __eq__(self, other):
        return (
            self.group == other.group
            and self.name == other.name
            and self.version == other.version
        )

    def __ge__(self, other):
        if self.group != other.group:
            return self.group >= other.group
        if self.name != other.name:
            return self.name >= other.name
        if self.version != other.version:
            return self.version >= other.version

    def __hash__(self):
        # needed because otherwise would differ in subclass,
        # causing problems for equality based on __hash__
        return hash((self.group, self.name, self.version))

    def __str__(self) -> str:
        return self.json()  # no indent (in contrast to base model)

    def supports(self, other: PluginRef) -> bool:
        """Return whether this plugin supports objects marked by given reference.

        True iff the package name, plugin group and plugin name agree,
        and the package of this reference has equal or larger minor version.
        """
        if self.group != other.group:
            return False
        if self.name != other.name:
            return False
        if self.version[0] != other.version[0]:  # major
            return False
        if self.version[1] < other.version[1]:  # minor
            return False
        return True

    @classmethod
    def _subclass_for(cls, group: NonEmptyStr):
        """Create a subclass of PluginRef with group field pre-set."""
        return create_model(f"PG{group.capitalize()}.PluginRef", __base__=cls, group=(Literal[group], group))  # type: ignore


class PluginBase(BaseModelPlus):
    """All Plugin inner classes must be called `Plugin` and inherit from this class."""

    group: ClassVar[str] = ""  # auto-set during plugin group init

    # for type checking (mirrors Fields)
    name: str
    version: SemVerTuple
    requires: List[PluginRef] = []

    def ref(self, *, version: Optional[SemVerTuple] = None):
        from ..plugins import plugingroups

        assert self.group, "Must be called from a subclass that has group set!"
        return plugingroups[self.group].PluginRef(
            name=self.name, version=version or self.version
        )

    def plugin_string(self):
        return f"metador.{self.group}.{self.name}.{to_semver_str(self.version)}"

    def __str__(self) -> str:
        # pretty-print semver in user-facing representation
        dct = dict(
            group=self.group, name=self.name, version=to_semver_str(self.version)
        )
        dct.update(
            self.json_dict(exclude_defaults=True, exclude={"name", "group", "version"})
        )
        return json.dumps(dct, indent=2)

    @classmethod
    def parse_info(cls, info, *, ep_name: str = ""):
        if isinstance(info, cls):
            return info  # nothing to do, already converted info class to PluginBase (sub)model

        expected_ep_name = f"{info.name}__{to_semver_str(info.version)}"
        ep_name = ep_name or expected_ep_name
        if ep_name != expected_ep_name:
            msg = f"{ep_name}: Based on plugin info, entrypoint must be called '{expected_ep_name}'!"
            raise ValueError(msg)
        try:
            fields = ChainMap(
                *(c.__dict__ for c in info.__mro__)
            )  # this will treat inheritance well
            # validate
            public_attrs = {k: v for k, v in fields.items() if is_public_name(k)}
            return cls(**public_attrs)
        except ValidationError as e:
            raise TypeError(f"{ep_name}: {ep_name}.Plugin validation error: \n{str(e)}")


PkgPlugins = Dict[str, List[PluginRef]]
"""Dict from plugin group name to plugins provided by a package."""


class PluginPkgMeta(MetadataSchema):
    """Metadata of a Python package containing Metador plugins."""

    name: NonEmptyStr
    """Name of the Python package."""

    version: SemVerTuple
    """Version of the Python package."""

    repository_url: Optional[AnyHttpUrl] = None
    """Python package source location (pip-installable / git-clonable)."""

    plugins: PkgPlugins = {}

    @classmethod
    def for_package(cls, package_name: str) -> PluginPkgMeta:
        """Extract metadata about a Metador plugin providing Python package."""
        # avoid circular import by importing here
        from importlib_metadata import distribution

        from ..plugin.entrypoints import DistMeta, distmeta_for
        from ..plugin.types import EPName, from_ep_name

        dm: DistMeta = distmeta_for(distribution(package_name))

        plugins: PkgPlugins = {}
        for group, ep_names in dm.plugins.items():
            plugins[group] = []
            for ep_name in ep_names:
                name, version = from_ep_name(EPName(ep_name))
                ref = PluginRef(group=group, name=name, version=version)
                plugins[group].append(ref)

        return cls(
            name=dm.name,
            version=dm.version,
            repository_url=dm.repository_url,
            plugins=plugins,
        )



"""Wrappers and helpers for RDFlib."""
from typing import Any, Dict, Union

import rdflib
import wrapt
from rdflib.term import Identifier

# NOTE: could evaluate oxrdflib if rdflib turns out being too slow


def at_most_one(g):
    """Return at most one element from a generator.

    If there are more, will raise ValueError.
    """
    ret = next(g, None)
    if next(g, None) is None:
        return ret
    else:
        raise ValueError()


class GraphNode(wrapt.ObjectProxy):
    """Wrapper for rdflib helping to navigate entities in an `rdflib.Graph`."""

    _self_graph: rdflib.Graph
    obj: Identifier

    def __init__(self, graph: rdflib.Graph, obj: Identifier = None):
        super().__init__(obj)
        self._self_graph = graph

    @property
    def graph(self):
        return self._self_graph

    @property
    def node(self):
        return self.__wrapped__

    def wrap(self, obj: Identifier):
        return GraphNode(self.graph, obj)

    def is_literal(self):
        return isinstance(self.__wrapped__, rdflib.Literal)

    def is_uriref(self):
        return isinstance(self.__wrapped__, rdflib.URIRef)

    # ----

    def edges_in(self):
        return map(self.wrap, self.graph.predicates(object=self.node, unique=True))  # type: ignore

    def edges_out(self):
        return map(self.wrap, self.graph.predicates(subject=self.node, unique=True))  # type: ignore

    def objects(self, pred):
        return map(self.wrap, self.graph.objects(self.node, pred, unique=True))  # type: ignore

    def subjects(self, pred):
        return map(self.wrap, self.graph.subjects(pred, self.node, unique=True))  # type: ignore

    def subject(self, pred):
        try:
            return at_most_one(self.subjects(pred))
        except ValueError:
            msg = f"Expected to get exactly one match for: (*, {pred}, {self.obj})"
            raise ValueError(msg)

    def object(self, pred):
        try:
            return at_most_one(self.objects(pred))
        except ValueError:
            msg = f"Expected to get exactly one match for: ({self.obj}, {pred}, *)"
            raise ValueError(msg)


class RDFParser(wrapt.ObjectProxy):
    """Helper wrapper to access entity properties backed by an RDFlib graph.

    Ensures that queries are only performed when needed (more efficient).

    Use this e.g. to parse linked data with expected structure into your custom
    data structures, e.g. for use in schemas.
    """

    __wrapped__: GraphNode

    def __init__(self, node: GraphNode):
        super().__init__(node)
        self._self_parsed: Dict[str, Any] = {}

    def __getattr__(self, key: str):
        # special case: return the method
        if key.startswith("parse_"):
            return wrapt.ObjectProxy.__getattr__(self, key)

        # try to return the attribute
        if val := self._self_parsed.get(key):
            return val
        # need first to compute the attribute
        if pfunc := getattr(self, f"parse_{key}"):
            val = pfunc(self.__wrapped__)
            self._self_parsed[key] = val
            return val
        # no parser defined -> pass through
        return getattr(self.__wrapped__, key)

    @classmethod
    def from_graph(cls, g: rdflib.Graph, obj: Union[str, rdflib.URIRef]):
        return cls(GraphNode(g, rdflib.URIRef(obj)))



import pytest

from metador_core.container import MetadorContainer
from metador_core.container.drivers import get_driver_type
from metador_core.container.utils import METADOR_VERSION_PATH
from metador_core.container.wrappers import (
    NodeAcl,
    UnsupportedOperationError,
    WrappedAttributeManager,
)


@pytest.fixture
def bibmeta_example(schemas):
    """Return a BibMeta instance.

    BibMeta is used here, because it is:
    * commonly relevant
    * actually using schema inheritance
    """
    BibMeta = schemas.get("core.bib", (0, 1, 0))
    Person = BibMeta.Fields.author.schemas.Person
    obj = BibMeta(
        name="Dataset1",
        abstract="Some longer description text.",
        dateCreated="2023-01-23",
        author=[Person(name="Jane Doe")],
        creator=Person(name="Jane Doe"),
    )
    return obj


def test_container_init(tmp_mc_path, mc_driver):
    drv_cls = mc_driver.value
    with MetadorContainer(tmp_mc_path, "w", driver=drv_cls) as m:
        c_uuid = m.metador.container_uuid
        c_src = m.metador.source

        # as expected
        assert m["/"]._self_container is m
        assert m.metador.spec_version[0] == 1
        assert m.metador.driver == drv_cls
        assert m.metador.driver_type == mc_driver
        assert m.mode == "r+"

    # can open based on retrieved "data source object"
    with MetadorContainer(c_src, "r", driver=drv_cls) as m:
        assert m.metador.container_uuid == c_uuid  # UUID did not change
        assert m.mode == "r"

    # can also just wrap a raw container that is already opened
    with drv_cls(c_src, "r") as r:
        assert MetadorContainer(r).__wrapped__ is r  # unpacked container object


def test_container_init_fail(tmp_mc_path, mc_driver):
    driver = mc_driver.value
    with driver(tmp_mc_path, "w") as r:
        # create empty container file and close it
        ...

    with pytest.raises(ValueError):
        # read-only and non-metador -> can't init
        with driver(tmp_mc_path, "r") as r:
            MetadorContainer(r)

    with MetadorContainer(driver(tmp_mc_path, "w")) as m:
        # init container, once writable, once only readable
        assert m.metador.container_uuid
        assert m.metador.spec_version[0] == 1

    with driver(tmp_mc_path, "w") as r:
        # mark with invalid version
        r[METADOR_VERSION_PATH] = "2.0"

    with pytest.raises(ValueError):
        # invalid spec version is assigned
        MetadorContainer(driver(tmp_mc_path, "r"))


# ----


def test_container_acl(tmp_mc_path, mc_driver):
    """Test soft ACL features of node wrappers."""
    drv_cls = mc_driver.value
    with MetadorContainer(tmp_mc_path, "w", driver=drv_cls) as m:
        m["foo/bar"] = [1, 2, 3]
        m["foo/bar"].attrs["hello"] = 123

        g = m["/"]
        g.restrict(read_only=True, local_only=True)
        g.restrict(read_only=False)  # should have no effect (can only set)
        assert g.acl[NodeAcl.read_only] == True

        # check that ACL is inherited downward
        g2 = g["foo"]
        assert g2.acl[NodeAcl.read_only] == True
        assert g2.acl[NodeAcl.local_only] == True
        # check that accessing parent does not allow to escape acl
        assert g2.parent.name == g.name
        assert g2.parent.acl == g.acl

        # narrow down local scope
        g2.restrict(local_only=True)

        # test local_only restriction:
        # can't access container or parent group
        with pytest.raises(UnsupportedOperationError):
            g2.file
        with pytest.raises(UnsupportedOperationError):
            g2.parent
        with pytest.raises(ValueError):
            g2["/"]  # absolute paths forbidden

        # can go down and back up to current level (local_only)
        d = g2["bar"]
        assert d.parent.name == g2.name
        assert d.parent.acl == g2.acl  # acl still not lost

        # test read_only restriction:
        with pytest.raises(UnsupportedOperationError):
            # can't add new dataset
            g2["qux"] = 123
        with pytest.raises(UnsupportedOperationError):
            # can't remove dataset
            del g2["bar"]
        with pytest.raises(UnsupportedOperationError):
            # can't edit in dataset
            d[0] = [4, 5, 6]
        with pytest.raises(UnsupportedOperationError):
            d.resize(1, 2, 3)

        # test attribute wrapper
        assert d.acl[NodeAcl.read_only]
        a = d.attrs
        assert isinstance(a, WrappedAttributeManager)
        assert a._self_acl == d.acl

        assert "hello" in a
        assert a["hello"] == 123
        assert a.get("hello") == 123
        assert list(a.keys()) == ["hello"]

        with pytest.raises(UnsupportedOperationError):
            # can't add attribute
            d.attrs["something"] = "new"
        with pytest.raises(UnsupportedOperationError):
            # can't remove attribute
            del d.attrs["hello"]

        # test skel_only on dataset and attributes
        d.restrict(skel_only=True)
        with pytest.raises(UnsupportedOperationError):
            d[()]  # cannot read data
        a = d.attrs
        assert "hello" in a
        assert list(iter(a)) == ["hello"]
        assert list(a.keys()) == ["hello"]
        with pytest.raises(UnsupportedOperationError):
            a["hello"]  # cannot read attribute
        with pytest.raises(UnsupportedOperationError):
            a.get("hello")
        with pytest.raises(UnsupportedOperationError):
            a.values()
        with pytest.raises(UnsupportedOperationError):
            a.items()

        # even with all restrictions enabled,
        # can still access metador specific properties
        d.meta
        d.metador


def test_container_toc_infos(tmp_mc_path, mc_driver):
    drv_cls = mc_driver.value
    with MetadorContainer(tmp_mc_path, "w", driver=drv_cls) as m:
        assert m.metador.driver == drv_cls
        assert m.metador.driver_type == get_driver_type(m.__wrapped__)
        assert m.metador.source is not None

        assert m.metador.container_uuid is not None
        assert m.metador.spec_version == [1, 0]


def test_node_metadata(tmp_mc_path, mc_driver, schemas, bibmeta_example):
    """Check that basic operations on node metadata work correctly."""
    meta = bibmeta_example
    BibMeta = type(meta)
    BibMeta_ref = BibMeta.Plugin.ref()
    DirMeta = schemas.get(schemas.parent_path(BibMeta)[-2])
    meta2 = meta.copy(update=dict(name="Dataset2"))
    assert meta != meta2

    drv_cls = mc_driver.value
    with MetadorContainer(tmp_mc_path, "w", driver=drv_cls) as m:
        m["foo/bar"] = [1, 2, 3]
        ds = m["foo/bar"]

        # try accessing non-existing metadata
        with pytest.raises(KeyError):
            ds.meta["core.bib"]

        # add metadata object
        assert len(ds.meta) == 0
        ds.meta["core.bib"] = meta
        assert len(ds.meta.items()) == 1

        # try to overwrite
        with pytest.raises(ValueError):
            ds.meta["core.bib"] = meta

        # get expected data back
        ret1 = ds.meta["core.bib"]
        assert isinstance(ret1, BibMeta)
        assert ret1.name == "Dataset1"

        # check contains (it has "inheritance semantics")
        assert "" not in ds.meta
        assert "something" not in ds.meta
        assert "core.bib" in ds.meta
        assert "core.dir" in ds.meta  # by inheritance (!!!)

        # get data back as parent schema (!!!)
        ret1b = ds.meta["core.dir"]
        assert isinstance(ret1b, DirMeta)
        assert ret1b.name == "Dataset1"

        # add another, actual instance, for the parent schema
        # that will take precedence over the "coerced" instance
        ds.meta["core.dir"] = meta2
        assert len(ds.meta.values()) == 2
        # get expected metadata back
        ret2 = ds.meta.get("core.dir")
        assert isinstance(ret1, DirMeta)
        assert ret2.name == "Dataset2"

        # query by schema name and version, check that hits are as expected

        # try passing no schema, plugin reference or schema class
        assert len(list(ds.meta.query())) == 2  # anything goes
        assert list(ds.meta.query(BibMeta_ref)) == [BibMeta_ref]
        assert len(list(ds.meta.query(BibMeta))) == 1
        # ask for any version
        assert len(list(ds.meta.query("core.bib"))) == 1
        assert len(list(ds.meta.query("core.dir"))) == 2
        # ask for a newer minor version - should return the older instances
        assert len(list(ds.meta.query("core.bib", (0, 2, 0)))) == 1
        assert len(list(ds.meta.query("core.dir", (0, 2, 0)))) == 2
        # ask for a newer major version - incompatible -> no hits
        assert len(list(ds.meta.query("core.bib", (1, 0, 0)))) == 0
        assert len(list(ds.meta.query("core.dir", (1, 0, 0)))) == 0

        # delete metadata -> should be gone
        del ds.meta["core.bib"]
        assert len(ds.meta.keys()) == 1
        assert len(list(iter(ds.meta))) == 1
        assert "core.bib" not in ds.meta
        assert "core.dir" in ds.meta

        del ds.meta["core.dir"]
        assert next(ds.meta.query(), None) is None
        assert "core.dir" not in ds.meta


def test_toc_metadata_schemas_packages(tmp_mc_path, mc_driver, bibmeta_example):
    """Check that schema and package tracking works correctly."""
    meta = bibmeta_example
    BibMeta = type(meta)
    BibMeta_ref = BibMeta.Plugin.ref()
    meta2 = meta.copy(update=dict(name="Dataset2"))
    assert meta != meta2

    drv_cls = mc_driver.value
    with MetadorContainer(tmp_mc_path, "w", driver=drv_cls) as m:
        m["foo/bar"] = [1, 2, 3]

        assert len(m.metador.schemas) == 0  # nothing should be here
        # attach some metadata with a non-trivial schema using inheritance:
        m["foo/bar"].meta["core.bib"] = meta

        # schema + providing package registered. check it is as expected
        assert len(m.metador.schemas) == 1  # core.bib is added
        assert BibMeta_ref in m.metador.schemas
        assert isinstance(m.metador.schemas[BibMeta_ref], dict)  # embedded json schema
        assert m.metador.schemas.provider(BibMeta_ref).name == "metador-core"

        # embedded schema parents/children interface (important for granular queries)
        bm_pp = m.metador.schemas.parent_path(BibMeta)
        assert len(bm_pp) == 2  # [core.dir ref, core.bib ref]
        assert bm_pp[-1] == BibMeta_ref
        assert bm_pp[-2].name == "core.dir"
        assert BibMeta_ref in m.metador.schemas.children("core.dir", (0, 1, 0))

        # add another similar metadata entry
        m["foo/qux"] = [4, 5, 6]
        m["foo/qux"].meta["core.dir"] = meta  # store explicitly as parent schema
        assert len(m.metador.schemas) == 2  # core.dir is added

        # removes nodes (should also kill metadata and clean up TOC)
        del m["foo/qux"]
        assert len(m.metador.schemas) == 1
        assert len(m.metador.schemas.packages) == 1
        del m["foo/bar"]
        assert len(m.metador.schemas) == 0
        assert len(m.metador.schemas.packages) == 0


def test_container_queries(tmp_mc_path, mc_driver, bibmeta_example):
    """Check that container/group-wide query works correctly."""
    meta = bibmeta_example
    BibMeta = type(meta)
    BibMeta_ref = BibMeta.Plugin.ref()
    meta2 = meta.copy(update=dict(name="Meta2"))
    assert meta != meta2

    drv_cls = mc_driver.value
    with MetadorContainer(tmp_mc_path, "w", driver=drv_cls) as m:
        # add some (meta)data
        m["foo/bar"] = [1, 2, 3]
        m["foo/bar"].meta["core.bib"] = meta

        # try container-wide queries
        assert len(list(m.metador.query("not_existing"))) == 0

        assert len(list(m.metador.query(BibMeta))) == 1
        assert len(list(m.metador.query(BibMeta_ref))) == 1
        assert len(list(m.metador.query("core.bib"))) == 1
        assert len(list(m.metador.query("core.bib", (0, 2, 0)))) == 1
        assert len(list(m.metador.query("core.bib", (1, 0, 0)))) == 0

        assert len(list(m.metador.query("core.dir"))) == 1
        assert len(list(m.metador.query("core.dir", (0, 2, 0)))) == 1
        assert len(list(m.metador.query("core.dir", (1, 0, 0)))) == 0

        # explicitly add parent schema
        m["foo/qux"] = [4, 5, 6]
        m["foo/qux"].meta["core.dir"] = meta2
        # create an empty group just for testing purposes
        m.create_group("baz")

        assert len(list(m.metador.query("core.dir"))) == 2
        assert len(list(m.metador.query("core.dir", (0, 2, 0)))) == 2

        assert len(list(m["baz"].metador.query("core.dir"))) == 0
        assert len(list(m["foo"].metador.query("core.dir"))) == 2
        assert len(list(m["foo/qux"].metador.query("core.bib"))) == 0
        assert len(list(m["foo/bar"].metador.query("core.bib"))) == 1


def test_group_operations_metadata_correct(tmp_mc_path, mc_driver, bibmeta_example):
    """Check that groups and datasets are moved together with metadata."""
    meta = bibmeta_example
    drv_cls = mc_driver.value
    with MetadorContainer(tmp_mc_path, "w", driver=drv_cls) as m:
        # create a group with dataset, both with a metadata entry
        # -> non-trivial case, covering relevant differences and most cases
        m["foo/bar"] = [1, 2, 3]
        m["foo/bar"].meta["core.bib"] = meta  # dataset metadata
        m["foo"].meta["core.bib"] = meta  # group metadata

        # sanity check
        expected = {"/foo", "/foo/bar"}
        assert set(map(lambda x: x.name, m.metador.query("core.bib"))) == expected
        assert len(m.metador._links.find_broken()) == 0  # check TOC was fine before
        assert len(m.metador._links.find_missing(m["/"])) == 0

        # ----
        # perform move and copy operations on both dataset and group nodes with metadata

        # copy dataset:
        m.copy("foo/bar", "foo/baz")
        expected = {"/foo", "/foo/bar", "/foo/baz"}
        assert set(map(lambda x: x.name, m.metador.query("core.bib"))) == expected
        # move dataset:
        m.move("foo/bar", "qux")
        expected = {"/foo", "/foo/baz", "/qux"}
        assert set(map(lambda x: x.name, m.metador.query("core.bib"))) == expected
        # copy and move a group:
        m.copy("foo", "blub")
        m.move("foo", "bla")
        expected = {"/bla", "/bla/baz", "/blub", "/blub/baz", "/qux"}
        assert set(map(lambda x: x.name, m.metador.query("core.bib"))) == expected

        # check that all metadata node links in TOC are still functional
        assert len(m.metador._links.find_missing(m["/"])) == 0
        assert len(m.metador._links.find_broken()) == 0

        uuids = set()
        for n_path in expected:
            # check that the metadata can actually be accessed
            assert m[n_path].meta.get("core.bib") is not None

            # check that uuid mangling worked for move/copy (agrees with dataset node path)
            meta_info = m[n_path].meta._objs["core.bib"]
            assert meta_info.uuid == type(meta_info).from_node(meta_info.node).uuid
            uuids.add(meta_info.uuid)  # collect
        # check that the metadata objects are all actually distinct
        assert len(uuids) == len(expected)

        # ----
        # remove and see everything is really gone

        del m["/qux"]  # dataset
        del m["/bla"]  # group
        del m["/blub"]  # group

        # no data and metadata should be found now!
        assert len(list(m.keys())) == 0
        assert set(map(lambda x: x.name, m.metador.query("core.bib"))) == set()
        # no metadata node links in TOC or container should be left
        assert len(m.metador._links.find_missing(m["/"])) == 0
        assert len(m.metador._links.find_broken()) == 0



"""Tests for MetadorContainer."""
# from pathlib import Path

# import h5py
# from overrides import overrides
import pytest

# from metador_core.container import MetadorContainer
# from metador_core.hashutils import DEF_HASH_ALG, dir_hashsums
# from metador_core.ih5.container import IH5Record
# from metador_core.packer import DirDiff, Packer
from metador_core.packer.utils import DirValidationErrors, check_metadata_file
from metador_core.schema.common import FileMeta

pytest.skip(allow_module_level=True)  # TODO


def test_errors_add_append():
    errs = DirValidationErrors()
    assert not errs

    errs.add("a", "entry1")
    errs.add("a", "entry2")
    assert errs
    assert errs.errors["a"] == ["entry1", "entry2"]

    errs2 = DirValidationErrors()
    errs2.add("a", "entry3")
    errs2.add("a", "entry4")
    errs2.add("b", "entry5")
    assert errs2.errors["a"] == ["entry3", "entry4"]
    assert errs2.errors["b"] == ["entry5"]

    errs3 = DirValidationErrors()
    errs3.add("b", "entry6")

    errs.update(errs2, errs3)
    assert errs.errors["a"] == ["entry1", "entry2", "entry3", "entry4"]
    assert errs.errors["b"] == ["entry5", "entry6"]


def test_check_file(tmp_ds_path):
    tmp_ds_path.mkdir()
    assert tmp_ds_path.is_dir()

    # not existing
    file = tmp_ds_path / "test.yaml"

    assert not check_metadata_file(file)
    assert not check_metadata_file(file, schema=FileMeta)

    ret = check_metadata_file(file, required=True)
    assert ret.errors[str(file)][0].find("file not found") >= 0

    # exists, invalid
    with open(file, "w") as f:
        f.write("{ # }")

    assert not check_metadata_file(file, required=True)
    ret = check_metadata_file(file, schema=FileMeta)
    assert ret.errors[str(file)][0].find("validation error") >= 0

    # exists + valid
    with open(file, "w") as f:
        f.write(FileMeta(filename="test.yaml", hashsum="sha256:123").yaml())
    assert not check_metadata_file(file, schema=FileMeta)


# class DummyPacker(Packer):
#     """Do-nothing packer testing packing api."""

#     fail_check_directory = False
#     ex_check_directory = False
#     ex_pack_directory = False

#     @classmethod
#     @overrides
#     def check_dir(cls, data_dir: Path) -> DirValidationErrors:
#         if cls.ex_check_directory:
#             raise Exception("check_directory failed.")

#         if cls.fail_check_directory:
#             return DirValidationErrors({"error": ["check_directory"]})
#         return DirValidationErrors()

#     @classmethod
#     @overrides
#     def pack(
#         cls, data_dir: Path, diff: DirDiff, record: IH5Record, fresh: bool
#     ):
#         if cls.ex_pack_directory:
#             raise ValueError("pack_directory failed.")

#         # just count number of patches (base container is 0)
#         if "updates" not in record.attrs:
#             record.attrs["updates"] = 0
#         else:
#             if not isinstance(record.attrs["updates"], h5py.Empty):
#                 record.attrs["updates"] = record.attrs["updates"] + 1



import pytest
from pydantic import ValidationError

from metador_core.schema import MetadataSchema
from metador_core.schema.ld import LDIdRef, LDOrRef, LDSchema, ld, ld_decorator


def test_ld():
    @ld(context="https://hello.com", type="Dummy")
    class DummySchema(MetadataSchema):
        foo: str

    obj = DummySchema(foo="test")
    assert obj.dict() == {
        "@context": "https://hello.com",
        "@type": "Dummy",
        "foo": "test",
    }

    @ld(type="Dummy2")
    class DummySchema2(DummySchema):
        bar: int

    obj2 = DummySchema2(foo="test", bar=123)
    assert obj2.dict() == {
        "@context": "https://hello.com",
        "@type": "Dummy2",
        "foo": "test",
        "bar": 123,
    }

    @ld(context="https://world.com", blub="bla")
    class DummySchema3(DummySchema2):
        ...

    obj3 = DummySchema3(foo="test", bar=123)
    assert obj3.dict() == {
        "@blub": "bla",
        "@context": "https://world.com",
        "@type": "Dummy2",
        "foo": "test",
        "bar": 123,
    }


def test_ld_no_override():
    @ld(context="https://hello.com", type="Dummy")
    class DummySchema(MetadataSchema):
        foo: str

    # overriding -> raises exception

    with pytest.raises(ValueError):

        @ld(type="Dummy2", override=False)
        class DummySchema2(DummySchema):
            bar: int

    with pytest.raises(ValueError):

        @ld(context="https://override.com", override=False)
        class DummySchema2(DummySchema):  # noqa: F811
            bar: int

    # not overriding -> fine
    @ld(blub="value", override=False)
    class DummySchema2(DummySchema):  # noqa: F811
        bar: int


def test_custom_ld():
    myld = ld_decorator(context="https://hello.com", type="Dummy")

    # works even without arguments
    @myld
    class DummySchema(MetadataSchema):
        foo: str

    obj = DummySchema(foo="test")
    assert obj.dict() == {
        "@context": "https://hello.com",
        "@type": "Dummy",
        "foo": "test",
    }

    @myld(type="Dummy2")
    class DummySchema2(DummySchema):
        bar: int

    obj2 = DummySchema2(foo="test", bar=123)
    assert obj2.dict() == {
        "@context": "https://hello.com",
        "@type": "Dummy2",
        "foo": "test",
        "bar": 123,
    }

    @myld(context="https://world.com", blub="bla")
    class DummySchema3(DummySchema2):
        ...

    obj3 = DummySchema3(foo="test", bar=123)
    assert obj3.dict() == {
        "@blub": "bla",
        "@context": "https://world.com",
        "@type": "Dummy",  # <- presets ALWAYS override
        "foo": "test",
        "bar": 123,
    }


# ----


def test_ldschema():
    class DummySchema(LDSchema):
        foo: str

    obj = DummySchema(foo="test")
    assert obj.dict().get("id_") is None
    assert obj.dict().get("@id") is None

    assert not obj.is_ld_ref
    with pytest.raises(ValueError):
        assert obj.ref()

    obj.id_ = "fancy_id"
    assert obj.dict().get("id_") is None
    assert obj.dict()["@id"] == "fancy_id"
    o_ref = obj.ref()

    assert o_ref.is_ld_ref
    assert o_ref.ref() is o_ref
    assert isinstance(o_ref, LDIdRef)
    assert isinstance(o_ref, LDSchema)
    assert o_ref.dict().get("id_") is None
    assert o_ref.dict()["@id"] == "fancy_id"

    # an ID ref has an id_ field
    with pytest.raises(ValidationError):
        LDIdRef(id_="")
    with pytest.raises(ValidationError):
        LDIdRef()


def test_ld_or_ref():
    class DummySchema(LDSchema):
        foo: str

    class OuterSchema(LDSchema):
        bar: LDOrRef[DummySchema]  # either LDIdRef or a DummySchema

    # all of these work
    OuterSchema(bar=LDIdRef(id_="my_id"))
    OuterSchema(bar=DummySchema(foo="blub"))
    OuterSchema(bar=DummySchema(foo="blub", id_="my_id"))

    # these do not
    with pytest.raises(ValidationError):
        OuterSchema()
    with pytest.raises(ValidationError):
        OuterSchema(bar=5)



"""Support for dynamically registered JSON encoders in pydantic.

Preparation:

For your top level BaseModel, set `DynEncoderModelMeta` as metaclass, e.g.

```
class MyBaseModel(BaseModel, metaclass=DynEncoderModelMeta):
    ...
```

(If you already use a custom metaclass for your base model,
add the `DynJsonEncoder` metaclass mixin)

Usage:

Decorate any class with `@json_encoder(ENCODER_FUNCTION)`.

To add an encoder for some existing class you cannot decorate,
use `add_json_encoder(ClassName, func)`.

Note that `json_encoders` declared as intended by Pydantic in the `Config` section
will always be prioritized over the dynamic encoder. This means, that the dynamic
encoders are only triggered for classes that are not models themselves
(because pydantic handles them already).

Also note that to prevent bugs, you cannot override encoders for a class that already
has a registered dynamic encoder. Use the normal pydantic mechanisms for
cases where this is really needed.

Ideally, design your classes in a way that there is a 1-to-1 relationship between
class and desired JSON encoder (e.g. you can have different subclasses with
different encoders).
"""

from typing import Any, Callable, Dict, Type

from pydantic.main import ModelMetaclass

_reg_json_encoders: Dict[Type, Callable[[Any], str]] = {}
"""Global registry of declared JSON encoders."""


def json_encoder(func):
    """Decorate a class to register a new JSON encoder for it."""

    def reg_encoder(cls):
        if issubclass(cls.__class__, ModelMetaclass):
            raise TypeError("This decorator does not work for pydantic models!")
        if hasattr(cls, "__dataclass_fields__"):
            raise TypeError("This decorator does not work for dataclasses!")

        if cls in _reg_json_encoders:
            raise ValueError(f"A JSON encoder function for {cls} already exists!")

        _reg_json_encoders[cls] = func
        return cls

    return reg_encoder


def add_json_encoder(cls, func):
    """Register a JSON encoder function for a class."""
    return json_encoder(func)(cls)


# ----


def _dynamize_encoder(encoder_func):
    """Wrap the JSON encoder pydantic generates from the Config to support the dynamic registry."""

    def wrapped_encoder(obj):
        try:
            # try the default lookup
            return encoder_func(obj)
        except TypeError as e:
            if enc := _reg_json_encoders.get(type(obj)):
                return enc(obj)  # try dynamic lookup
            raise e

    return wrapped_encoder


class DynJsonEncoderMetaMixin(type):
    """Metaclass mixin to first look in dynamic encoder registry.

    Combine this with (a subclass of) `ModelMetaClass` and use it for your custom base model.
    """

    def __init__(self, name, bases, dct):
        super().__init__(name, bases, dct)
        self.__json_encoder__ = staticmethod(_dynamize_encoder(self.__json_encoder__))


class DynEncoderModelMetaclass(DynJsonEncoderMetaMixin, ModelMetaclass):
    """Set this metaclass for your custom base model to enable dynamic encoders."""



import pytest
from hypothesis import given
from hypothesis import strategies as st

from metador_core.plugin.types import (
    EPGroupName,
    EPName,
    SemVerStr,
    SemVerTuple,
    ep_name_has_namespace,
    from_ep_group_name,
    from_ep_name,
    from_semver_str,
    is_pluginlike,
    plugin_args,
    to_ep_group_name,
    to_ep_name,
    to_semver_str,
)


class A:
    pass


class B:
    class Plugin:
        pass


class C:
    class Plugin:
        group = "blub"
        name = "bla"
        version = "bli"


class Info:
    def __init__(self, a="a", b="b", c=(1, 2, 3)):
        self.group = a
        self.name = b
        self.version = c


class D:
    Plugin = Info()


def test_is_pluginlike():
    assert not is_pluginlike(A)  # no inner 'Plugin'
    assert not is_pluginlike(B)  # missing fields
    assert is_pluginlike(C)  # a class
    assert is_pluginlike(D)  # an instance


def test_plugin_args_simple():
    # first argument is string (just name)
    name, ver = plugin_args("")
    assert name == "" and ver is None

    name, ver = plugin_args("test", None)
    assert name == "test" and ver is None

    name, ver = plugin_args("test", (1, 2, 3))
    assert name == "test" and ver == (1, 2, 3)

    # should work
    name, ver = plugin_args("test", (1, 2, 3), require_version=True)
    # should fail (could not infer any version)
    with pytest.raises(ValueError):
        name, ver = plugin_args("test", None, require_version=True)


@pytest.mark.parametrize("obj", [Info(), D()])
def test_plugin_args_complex(obj):
    # name and version come both from first argument
    name, ver = plugin_args(obj)
    assert name == "b" and ver == (1, 2, 3)
    # the implicitly passed version should suffice
    name, ver = plugin_args(obj, require_version=True)
    assert name == "b" and ver == (1, 2, 3)
    # the explicitly passed version should override
    name, ver = plugin_args(obj, (3, 2, 1))
    assert name == "b" and ver == (3, 2, 1)


# ----


@pytest.mark.parametrize("s", ["", "a", "1", "1.0", "1.0.5.1", "1.-1.4"])
def test_semverstr_invalid(s):
    with pytest.raises(TypeError):
        SemVerStr(s)


@given(st.from_type(SemVerTuple))
def test_semverstr_conversion(obj):
    semver_str = to_semver_str(obj)
    assert SemVerStr(semver_str)
    assert from_semver_str(semver_str) == obj


@pytest.mark.parametrize("s", ["", "x_y", "x_1.0", "y_1.0.3.1", "x_1.0.0"])
def test_epname_invalid(s):
    with pytest.raises(TypeError):
        EPName(s)


def test_epname_namespace():
    assert ep_name_has_namespace(EPName("abc.def__0.1.0"))
    assert not ep_name_has_namespace(EPName("abc__1.2.3"))


@given(st.from_regex(EPName.__pattern__, fullmatch=True))
def test_epname_conversion(obj):
    assert EPName(obj)
    name, ver = from_ep_name(obj)

    # sanity check
    assert len(name) >= 2
    assert SemVerStr(to_semver_str(ver))

    # check inverses, but use normalized rep (0.00.01 -> 0.0.1)
    normalized = to_ep_name(name, ver)
    assert to_ep_name(*from_ep_name(normalized)) == normalized


@given(st.from_regex(EPGroupName.__pattern__, fullmatch=True))
def test_ep_group_name_conversion(obj):
    assert to_ep_group_name(from_ep_group_name(obj)) == obj



import json
from dataclasses import dataclass
from typing import Optional

import pytest
from pydantic import BaseModel, ValidationError

from metador_core.schema.encoder import (
    DynEncoderModelMetaclass,
    add_json_encoder,
    json_encoder,
)
from metador_core.schema.parser import BaseParser, ParserMixin

# ----
# Test custom parser


class SomeModel(BaseModel):
    x: str


class TrivialParsed(ParserMixin, SomeModel):
    """Parser does nothing, only accepts an instance."""

    Parser = BaseParser


class StringParsed(ParserMixin, SomeModel):
    """Parser packs a string into the model."""

    class Parser(BaseParser):
        schema_info = dict(title="Foobar")
        strict = True

        @classmethod
        def parse(cls, target, v):
            return target(x=v)


class SubModel(StringParsed):
    """Subclass of class with custom parser, should not inherit it."""


class InvalidParsed(ParserMixin, SomeModel):
    """Incorrect parser implementation (violates strict flag)."""

    class Parser(BaseParser):
        strict = True

        @classmethod
        def parse(cls, target, v):
            return v  # <- invalid, not returning target instance


class OuterModel(BaseModel):
    y: StringParsed
    s: Optional[SubModel]
    z: Optional[TrivialParsed]
    i: Optional[InvalidParsed]


def test_custom_parser():
    with pytest.raises(TypeError):

        class Invalid(ParserMixin, SomeModel):
            class Parser:
                ...  # <- not subclass of BaseParser

        class FaultyModel(BaseModel):
            x: Invalid

        list(FaultyModel.__get_validators__())

    with pytest.raises(RuntimeError):  # for i
        obj = OuterModel(y="hello", i={"x": "everyone"})

    with pytest.raises(ValidationError):  # for z
        obj = OuterModel(y="hello", z="world")

    with pytest.raises(ValidationError):  # for s
        obj = OuterModel(y="hello", s="everyone")

    # test parsing
    obj = OuterModel(y="hello", s=SubModel(x="test"), z=TrivialParsed(x="world"))
    assert isinstance(obj.y, StringParsed)
    assert isinstance(obj.z, TrivialParsed)

    # make sure the description is attached to jsonschema
    assert obj.schema()["properties"]["y"]["title"] == "Foobar"


# ----
# Test custom JSON encoder


def test_invalid_custom_encoder():
    with pytest.raises(TypeError):
        # not allowed on pydantic models
        @json_encoder(lambda _: "foo")
        class MyModel(BaseModel):
            ...

    with pytest.raises(TypeError):
        # not allowed on dataclasses
        @json_encoder(lambda _: "foo")
        @dataclass
        class MyModel2:
            ...

    @json_encoder(lambda _: "foo")
    class MyModel3:
        ...

    with pytest.raises(ValueError):
        # not allowed twice
        add_json_encoder(MyModel3, lambda _: "bar")


def test_custom_encoder():
    class CustomClass:
        ...

    @json_encoder(lambda _: "foo")
    class JsonableClass(CustomClass):
        ...

    class MyBaseModel(BaseModel, metaclass=DynEncoderModelMetaclass):
        class Config:
            arbitrary_types_allowed = True

    class DummyModel(MyBaseModel):
        x: JsonableClass

    # check that serialization is as expected (custom function)
    j = DummyModel(x=JsonableClass()).json()
    d = json.loads(j)
    assert d.get("x") == "foo"

    class NonJsonableClass:
        ...

    class DummyModel2(MyBaseModel):
        x: NonJsonableClass

    o = DummyModel2(x=NonJsonableClass())
    o.dict()  # should work
    with pytest.raises(TypeError):  # not JSON serializable!
        o.json()


# ----
# Test combination (non-pydantic class with added custom parser and json encoder)


def test_custom_parser_encoder():
    class CombinedBaseModel(BaseModel, metaclass=DynEncoderModelMetaclass):
        ...

    class CustomClass:
        ...

    @json_encoder(lambda _: "foo")
    class ParsableClass(ParserMixin, CustomClass):
        class Parser(BaseParser):
            strict = True

            @classmethod
            def parse(cls, target, val):
                if val == "foo":
                    return target()
                raise ValueError(f"Invalid input: {val}")

    class DummyModel2(CombinedBaseModel):
        x: ParsableClass

    with pytest.raises(ValidationError):
        j = DummyModel2(x="bar").json()

    # serialize, deserialize
    j = DummyModel2(x="foo").json()
    d = json.loads(j)
    o = DummyModel2.parse_obj(d)
    assert isinstance(o.x, ParsableClass)
    assert isinstance(o.x, CustomClass)

    # make serialization invalid -> should fail
    d["x"] = "bar"
    with pytest.raises(ValidationError):
        o = DummyModel2.parse_obj(d)
